# This code is part of qmatchatea.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

import pathlib
import unittest
import subprocess


class TestBlackFormatting(unittest.TestCase):
    def test_black_formatting(self):
        # Enter the following folders up to a defined depth
        folders_depth = {
            "qmatchatea": 2,
            "tests": 1,
        }

        # Get the root path of the repository based on the current file
        root = str(pathlib.Path(__file__).parent.parent.resolve())

        paths = [root + "/*.py"]
        for elem, depth in folders_depth.items():
            fill_path = "/"
            for ii in range(depth):
                paths.append(root + "/" + elem + fill_path + "*.py")

                fill_path += "*/"

        cmd_call = ["black", "--check", "--exclude=Examples"] + paths
        result = subprocess.call(" ".join(cmd_call), shell=True)
        success = int(result) == 0

        self.assertTrue(success, "Repository did not pass Black formatting.")



# This code is part of qmatchatea.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""
In order to simulate the quantum circuit we first need to preprocess it, to put it in a way
suitable for the MPS simulation. Afterwards, **if** a fortran backend is selected the parsing
process is applied, i.e. a suitable input file is generated for the fortran backend.
Thus, the two main processes explained here are:

1. Preprocessing of the circuit
2. Parsing the circuit for the Fortran I/O

Preprocessing
~~~~~~~~~~~~~

The tensor network ansatz that is used in the simulator is called Matrix Product States (MPS).
In this ansatz each degree of freedom (qubit, mode, particle) is treated as a rank-3 tensor.
We can only apply operators which are **local** or that applies to **nearest neighbors**.
This means that, when a circuit is passed to the preprocessing it is translated in an equivalent
circuit that follows these two properties. While for quantum optics these properties are already
ensured, for qubits quantum circuit we have to perform some transformations. We use qiskit for them,
and in particular we:

- Map the circuit using a pre-defined set of gates, called basis_gates.
  These gates can be passed to the function, to satisfy
  a particular constraint of a physical machine we want to simulate
- Map the circuit into a linear circuit. By an optimized application of *swap*
  gates we map non-local two-qubit gates into a series
  of local two-qubit gates.

These operations are automatically performed by the higher level function :py:func:`run_simulation`.
However, it is important to take into account that these operations require time, which is
proportional to the size of the circuit. For this reason, if the user has to perform multiple
simulation of the same circuit varing other parameres, e.g. the bond dimension, it is suggested to
perform the preprocessing only once, and then disable the mapping and the linearization of the
function.

.. warning::
    The preprocessing and linearization using qiskit modifies the order of the qubits.
    This is a behavior we don't want to experience, and such inside the preprocessing
    procedure the qubits are again ordered in the initial layout.

Parsing
~~~~~~~

These procedures are mirroring the Fortran procedure for the argument parsing. This means that,
if these are modified also the others **must** be modified and viceversa. However, the parsing
is not the only thing they do. Indeed, the quantum circuit from qiskit or the Program from
strawberry fields are analyzed, and the operators used inside them are saved into a file for
fortran under the form of tensors. The syntax that an input file must follow after the parsing
is the following:

.. code-block:: console

    total_num_sites
        total_num_classical_bits
        num_remaining_lines
        operator_name operator_idx num_sites
            first_site second_site
        operator_name operator_idx num_sites
            first_site

The *operator_idx* is a unique integer map to the operator tensor, stored in a different file.

"""


import os
import os.path
import warnings
from qiskit import QuantumCircuit, transpile, transpiler
import numpy as np
from qtealeaves import StrBuffer
from .sf_utils import optical_gates
from .qk_utils import get_index, qk_transpilation_params
from .tensor_compiler import tensor_compiler

try:
    from strawberryfields import Program

    SF_IS_IMPORTED = True
except ModuleNotFoundError:
    warnings.warn(
        "Not able to import strawberry-fields. Some functions will not be usable.",
        ImportWarning,
    )
    SF_IS_IMPORTED = False

__all__ = ["pre_parser", "preprocess"]
# Swap matrix
swap = np.array([[1, 0, 0, 0], [0, 0, 1, 0], [0, 1, 0, 0], [0, 0, 0, 1]])


# Strawberry fields gates are contained in optical_gates.
def _pre_parser_qk(qc, operators, folder_dst):
    """
    Map a qiskit quantum circuit into a pair of text files to be read by the fortran
    simulator. The first one contains all the tensors that will be applied on the MPS as
    tensor structure, while the second the circuit structure with the following syntax:

    .. code-block::

        total_num_qubits
        total_num_bits
        num_remaining_lines
        gate_name gate_idx num_qubits
            first_qubit second_qubit
        gate_name gate_idx first num_qubits
            first_qubit

    Parameters
    ----------
    qc: qiskit.QuantumCircuit
        Quantum circuit instance
    operators: :py:class:`QCOperators`
        Operator class with the observables operators ALREADY THERE
    folder_dst: string
        path to the folder where the operators file `TENSORS/operator.dat` and
        `circuit.dat` will be saved

    Returns
    -------
    num_qubits: int
        Number of qubits
    id_mapping: dict
        Mapping between operator name and integer ID
    circ: str
        String containing the circuit
    """
    assert os.path.isdir(folder_dst), "Provided PATH does not exists"

    data = qc.data  # data structure of the quantum circuit
    qubits = []  # List of target qubits
    clbits = []  # List of used classical bits
    temp = 0  # Temporary variable for the definition of parametric gates names
    op_names = []
    basic_names = []
    conditions = []

    for instance in data:
        gate_name = instance[0].name
        basic_names.append(gate_name)
        if gate_name == "barrier":
            gate_mat = np.zeros(1)
        elif gate_name == "measure":
            gate_mat = np.zeros(1)
        else:
            gate_mat = instance[0].to_matrix()
        num_qubits = len(instance[1])
        num_params = len(instance[0].params)

        qubits.append(np.array([get_index(qc, instance[1][0])]))
        clbits.append([get_index(qc, cl) for cl in instance[2]])
        conditions.append(instance[0].condition)

        # If the gate has not yet been encountered we add it to the
        # name dictionary and the tensor list. Notice that this happens
        # in two cases: if the gate_name is new OR if the gate is parametric,
        # since parametric gates have different matric representation depending on the parameters
        if gate_name not in operators.ops.keys() or num_params > 0:
            if num_params > 0:
                gate_name = gate_name + str(temp)
                temp += 1
            # the transposition is due to an operation done inside the operator class
            # where it stores the state column-major order
            if num_qubits == 2 and gate_name != "barrier":
                gate_mat = gate_mat.reshape(2, 2, 2, 2)
            operators.ops[gate_name] = gate_mat

        if num_qubits == 2:
            qubits[-1] = np.append(qubits[-1], get_index(qc, instance[1][1]))

        op_names.append(gate_name)

    # Save tensors on file and retrieve ID mapping
    _, id_mapping = operators.write_input_3(folder_dst)

    # Save circuit on file
    circuit_dst = os.path.join(folder_dst, "circuit.dat")
    circ = StrBuffer()
    circ.write(str(qc.num_qubits) + "\n")
    circ.write(str(qc.num_clbits) + "\n")
    circ.write(str(len(op_names)) + "\n")
    for jj, name in enumerate(op_names):
        # +1 to be consistent with the fortran standard, where arrays start from 1
        circ.write(
            basic_names[jj]
            + " "
            + str(id_mapping[name])
            + " "
            + str(len(qubits[jj]))
            + " "
            + str(len(clbits[jj]))
            + f" {int(conditions[jj] is not None)}"
            + "\n"
        )
        circ.write("\t")
        for qub in qubits[jj]:
            circ.write(str(qub + 1) + " ")  # Qubit numbers
        for clbit in clbits[jj]:
            circ.write(str(clbit + 1) + " ")  # Classical bit numbers
        # Index of the interested classical bit and supposed vakye
        if conditions[jj] is not None:
            circ.write(f"{get_index(qc, conditions[jj][0])} {int(conditions[jj][1])} ")
        circ.write("\n")

    with open(circuit_dst, "w") as fh:
        fh.write(circ())

    return qc.num_qubits, id_mapping, circ()


def _pre_parser_sf(sf_program, operators, folder_dst, fock_cutoff, **kwargs):
    """
    Map a strawberry_fields program into a pair of text files to be read by the fortran
    simulator. The first one contains all the tensors that will be applied on the MPS as
    tensor structure, while the second the circuit structure with the following syntax:

    .. code-block::

        total_num_modes
        0
        num_remaining_lines
        gate_name gate_idx num_modes
            first_mode second_mode
        gate_name gate_idx num_modes
            first_mode

    Parameters
    ----------
    qc: sf.Program
        Strawberry fields program
    operators: :py:class:`QCOperators`
        Operator class with the observables operators ALREADY THERE
    folder_dst: string
        path to the folder where the operators file `TENSORS/operator.dat` and
        `circuit.dat` will be saved
    fock_cutoff: int
        cutoff of the fock space

    Returns
    -------
    num_modes: int
        Number of modes
    id_mapping: dict
        Mapping between operator name and integer ID
    circ: str
        String containing the circuit
    """
    if not SF_IS_IMPORTED:
        raise ImportError("Strawberryfields library is not imported.")
    assert os.path.isdir(folder_dst), "Provided PATH does not exists"

    data = sf_program.circuit  # data structure of the program
    modes = []  # List of target qubits
    temp = 0  # Temporary variable for the definition of parametric gates names
    op_names = []
    basic_names = []

    for instance in data:
        gate_name = type(instance.op).__name__
        basic_names.append(gate_name)
        modes_temp = [idx.ind for idx in instance.reg]
        params = instance.op.p
        gate_mat = optical_gates[gate_name](*params, fock_cutoff)

        num_modes = len(modes_temp)
        num_params = len(params)

        modes.append(modes_temp)

        # If the gate has not yet been encountered we add it to the
        # name dictionary and the tensor list. Notice that this happens
        # in two cases: if the gate_name is new OR if the gate is parametric,
        # since parametric gates have different matric representation depending on the parameters
        if gate_name not in operators.ops.keys() or num_params > 0:
            if num_params > 0:
                gate_name = gate_name + str(temp)
                temp += 1

            if num_modes == 2:
                gate_mat = np.transpose(gate_mat, [0, 2, 1, 3])

            operators.ops[gate_name] = gate_mat

        op_names.append(gate_name)

    # Save tensors on file and retrieve ID mapping
    _, id_mapping = operators.write_input_3(folder_dst, **kwargs)

    # Save circuit on file
    circuit_dst = os.path.join(folder_dst, "circuit.dat")
    circ = StrBuffer()
    circ.write(str(sf_program.num_subsystems) + "\n")
    circ.write("0 \n")  # 0 classical bits.
    # In this way we are consistent with the qiskit preprocessor
    # and thus need only a single reader in Fortran
    circ.write(str(len(op_names)) + "\n")
    for jj, name in enumerate(op_names):
        # +1 to be consistent with the fortran standard, where arrays start from 1
        circ.write(
            basic_names[jj]
            + " "
            + str(id_mapping[name])
            + " "
            + str(len(modes[jj]))
            + " "
            + str(0)
            + "\n"
        )  # At the moment no classical bit is present for quantum optics
        circ.write("\t")
        for qub in modes[jj]:
            circ.write(str(qub + 1) + " ")  # Mode numbers
        circ.write("\n")

    with open(circuit_dst, "w") as fh:
        fh.write(circ())

    return sf_program.num_subsystems, id_mapping, circ()


def pre_parser(circ, operators, folder_dst, **kwargs):
    """
    Interface for the preparser of qiskit and strawberry fields

    Parameters
    ----------
    circ: strawberryfields.program.Program or qiskit.circuit.quantumcircuit.QuantumCircuit
        Quantum circuit instance
    operators: :py:class:`QCOperators`
        Operator class with the observables operators ALREADY THERE
    folder_dst: str
        path to the folder where the operators file `TENSORS/operator.dat` and
        `circuit.dat` will be saved
    **kwargs: int
        The fock cutoff to be provided if type(circ)=strawberryfields.program.Program
        sparse, if to use sparse tensors for quantum optics

    Returns
    -------
    n_sites: int
        Number of sites of the MPS
    operator_mapping: dict
        Mapping between operator name and integer ID
    circ_str: str
        String containing the circuit
    """
    assert os.path.isdir(folder_dst), "Provided PATH does not exists"

    if isinstance(circ, QuantumCircuit):
        n_sites, operator_mapping, circ_str = _pre_parser_qk(
            circ, operators, folder_dst
        )
    elif SF_IS_IMPORTED and isinstance(circ, Program):
        n_sites, operator_mapping, circ_str = _pre_parser_sf(
            circ, operators, folder_dst, **kwargs
        )
    else:
        raise TypeError(
            "Only qiskit quantum circuit of strawberry fields programs are implemented.\
            Your circuit is of type:"
            + str(type(circ))
        )

    return n_sites, operator_mapping, circ_str


def _bubble_sort(nums):
    """Given an integer array sort it by only swapping nearest neighbors elements,
        and return the necessary swaps to bring the array into sorted shape

    Parameters
    ----------
    nums : list of int
        Array of integer to be sorted in ascendant order

    Returns
    -------
    swaps : list of tuples of int
        Indexes of the qubits that has to be swapped
    """
    # We set swapped to True so the loop looks runs at least once
    swapped = True
    swaps = []
    while swapped:
        swapped = False
        for ii in range(len(nums) - 1):
            if nums[ii] > nums[ii + 1]:
                # Swap the elements
                nums[ii], nums[ii + 1] = nums[ii + 1], nums[ii]
                # Set the flag to True so we'll loop again
                swapped = True
                swaps.append((ii, ii + 1))
    return swaps


def _reorder_qk(linearized_qc):
    """Qiskit transpiler does not control the final layout of the qubits, i.e.
        we don't know if they are ordered correctly. It takes care of this problem
        by rearrenging the measurements on the classical register. However, it is
        a problem for the mps simulator. We so reorder the qubit register by applying
        swap gates.

    Parameters
    ----------
    linearized_qc : :py:class:`qiskit.QuantumCircuit`
        Circuit linearized trhough the qiskit transpiler

    Returns
    -------
    linearized_qc : :py:class:`qiskit.QuantumCircuit`
        Circuit with the correct ending qubits
    """
    num_qub = linearized_qc.num_qubits
    # Read the final layout by the rearrengement of the measurements
    final_perm = [0] * num_qub
    for operator, qubits, clbits in linearized_qc.data[
        linearized_qc.size() - num_qub + 1 :
    ]:
        if operator.name == "measure":
            final_perm[get_index(linearized_qc, qubits[0])] = get_index(
                linearized_qc, clbits[0]
            )
    # Get the combination of swaps necessary trhough the bubble_sort algorithm,
    # which is O(num_qub^2)
    swaps = _bubble_sort(final_perm)
    # Perform the swapping operations
    linearized_qc.remove_final_measurements()
    for ii, jj in swaps:
        linearized_qc.swap(ii, jj)

    return linearized_qc


def _preprocess_qk(qc, qk_params=qk_transpilation_params()):
    """Transpile the circuit to adapt it to the linear structure of the MPS, with the constraint
    of having only the gates basis_gates

     Parameters
    ----------
    qc: QuantumCircuit
         qiskit quantum circuit
    linearize: bool, optional
        If True use qiskit transpiler to linearize the circuit. Default to True.
    basis_gate: list, optional
        If not empty decompose using qiskit transpiler into basis gate set
    optimization: intger, optional
        Level of optimization in qiskit transpiler. Default to 3.

    Returns
    -------
    linear_qc: QuantumCircuit
        Linearized quantum circuit
    """
    # Empty circuit case
    if len(qc.data) == 0:
        return qc
    basis_gates = [] if qk_params.basis_gates is None else qk_params.basis_gates
    n_qub = qc.num_qubits
    linear_map = transpiler.CouplingMap.from_line(n_qub)
    # Assure that final measurements are present, by first eliminating and adding them
    qc.remove_final_measurements()
    qc.measure_all()
    # Transpile the circuit
    if len(basis_gates) > 0 and qk_params.linearize:
        linear_qc = transpile(
            qc,
            coupling_map=linear_map,
            optimization_level=qk_params.optimization,
            basis_gates=basis_gates,
            initial_layout=list(range(n_qub)),
        )
    elif len(basis_gates) == 0 and qk_params.linearize:
        linear_qc = transpile(
            qc,
            coupling_map=linear_map,
            optimization_level=qk_params.optimization,
            initial_layout=list(range(n_qub)),
        )
    elif len(basis_gates) > 0:
        linear_qc = transpile(
            qc,
            optimization_level=qk_params.optimization,
            basis_gates=basis_gates,
            initial_layout=list(range(n_qub)),
        )
    else:
        linear_qc = qc

    # Reorder the circuit
    linear_qc = _reorder_qk(linearized_qc=linear_qc)

    # Use the tensor compiler if requested
    if qk_params.tensor_compiler:
        linear_qc = tensor_compiler(linear_qc)

    return linear_qc


def _preprocess_sf(sf_program):
    """
    Preprocess a strawberry fields program, decomposing it.
    TODO

    Parameters
    ----------
    sf_program: strawberryfields.Program
            Input strawberryfields program

    Results
    -------
    sf_program_new: strawberryfields.Program
            Preprocessed strawberryfields program
    """
    if not SF_IS_IMPORTED:
        raise ImportError("Strawberryfields library is not imported.")
    sf_program_new = sf_program

    return sf_program_new


def preprocess(circ, **kwargs):
    """
        Interface for the preprocessing of qiskit and strawberry fields

        Parameters
        ----------
        circ: strawberryfields.program.Program or qiskit.circuit.quantumcircuit.QuantumCircuit
            Quantum circuit instance
        **kwargs: int
            linearize: bool, optional
                If True use qiskit transpiler to linearize the circuit. Default to True.
            basis_gate: list, optional
                If not empty decompose using qiskit transpiler into basis gate set
            optimization: intger, optional
                Level of optimization in qiskit transpiler. Default to 3.

        Returns
        -------
        preprocessed_circ: strawberryfields.program.Program or \
            qiskit.circuit.quantumcircuit.QuantumCircuit
            Preprocessed quantum circuit instance
    """

    if isinstance(circ, QuantumCircuit):
        preprocessed_circ = _preprocess_qk(circ, **kwargs)
    elif SF_IS_IMPORTED and isinstance(circ, Program):
        preprocessed_circ = _preprocess_sf(circ)
    else:
        raise TypeError(
            "Only qiskit quantum circuit of strawberry fields programs\
             are implemented. Your circuit is of type:"
            + str(type(circ))
        )

    return preprocessed_circ



# This code is part of qmatchatea.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""
General utility functions and classes for the simulation
"""

# Import necessary packages
import os
import pickle
import shutil
from warnings import warn
from datetime import datetime
from typing import OrderedDict
import numpy as np
from qtealeaves import write_nml
from qtealeaves.convergence_parameters import TNConvergenceParameters
from qtealeaves.emulator import MPS, TTN

from qtealeaves.observables import TNObservables

from .tn_utils import QCOperators, read_mps

EXE_PATH_DIR = os.path.dirname(__file__)
EXE_PATH = os.path.join(EXE_PATH_DIR, "bin/qmatchatea.exe")

__all__ = [
    "simulation_results",
    "print_state",
    "fidelity",
    "QCCheckpoints",
    "QCIO",
    "QCConvergenceParameters",
    "QCBackend",
    "SimpleHamiltonian",
]


class SimpleHamiltonian(dict):
    """
    Simple class for an Hamiltonian that extends a normal dictionary.
    The keys are the pauli strings, the values the coefficients.
    It is used for simplicity, since it has a `to_pauli_dict` method
    equivalent to qiskit and other methods to ease the construction.
    """

    def set_num_qubits(self, num_qubits):
        """
        Set the number of qubits the Hamiltonian is describing

        Parameters
        ----------
        num_qubits : int
            Number of qubits
        """
        self["num_qubits"] = num_qubits

    def add_term(self, hterms, qubits, coeff):
        """
        Add a term to the Hamiltonian acting on the
        qubits qubits. You do not need to specify the identities

        Parameters
        ----------
        hterms : str or array-like
            Pauli matrices to apply
        qubits : int or array-like
            Qubits where the terms acts
        coeff : complex
            Coefficient of the term

        Returns
        -------
        None
        """
        if np.isscalar(qubits):
            qubits = np.array([qubits])
            hterms = np.array([hterms])
        ordering = np.argsort(qubits)
        qubits = qubits[ordering]
        hterms = hterms[ordering]

        pauli_string = ""
        for hterm, qubit in zip(hterms, qubits):
            last_qubit = len(pauli_string)
            pauli_string += "I" * (qubit - last_qubit)
            pauli_string += hterm
        last_qubit = len(pauli_string)
        pauli_string += "I" * (self["num_qubits"] - last_qubit)

        self[pauli_string[::-1]] = coeff

    def to_pauli_dict(self):
        """
        Get the qiskit pauli dict representation, that can be later
        used in the observable class

        Returns
        -------
        dict
            dictionary with qiskit pauli_dict old format
        """
        pauli_dict = {"paulis": []}
        for key, val in self.items():
            if key == "num_qubits":
                continue

            pauli_dict["paulis"].append(
                {"label": key, "coeff": {"real": np.real(val), "imag": np.imag(val)}}
            )
        return pauli_dict


class simulation_results:
    """
    This class is used to handle the results of a simulation with qmatchatea.
    You can use it to:

    - Retrieve the results of a FORTRAN simulation;
    - Save the results of a simulation (see the `save_results()` method);
    - Load the results of a simulation (see the `load_results()` method);

    The results can be then accessed in python through the following properties.

    Input Properties
    ----------------
    date_time: str
        The date-time when the simulation started as ``Year-month-day-Hour:Minute:Second``
    input_params : dict
        Dictionary containing all the input informations
    initial_state : List[np.ndarray] or str
        The initial state of the simulation

    Output Properties
    -----------------
    fidelity: float
        Lower bound on the fidelity of the final state
    singular_values_cut : np.ndarray
        Singular values cut during the simulation
    measures: dict
        Projective measures of the simulation, in the format `{"000":15, "011":24}`
    measure_probabilities: List[dict or None]
        The 0th element is the result of the unbiased (OPES) sampling, with a format
        `{"000" : (0, 1)}` where the values are the left and right boundaries of the
        probability interval.
        The 1st element is even probability and the 2nd element greedy probabilities,
        with the format `{"000": 1}`, where the value is the probability of the state
    tens_net : List[np.ndarray]
        The tensor network state
    statevector: np.ndarray or None
        The statevector of the final state. Returned only if `num_sites<30`
    entanglement: dict or None
        Entanglement across bipartition, where the key is the bipartition. An example
        for a 2-site MPS system in a product state is `{(0,1) : 0}`
    observables : dict
        The dictionary containing ALL the observables of the simulation
    computational_time: float or None
        Time of the simulation
    """

    def __init__(self, input_params=None, observables=None):
        """
        Initialization. Provide a input params dictionary only if you want to retrieve
        the result directly from the Fortran working folder.
        Otherwise, if you want to load previous results, just initialize the class
        without parameters.

        Parameters
        ----------
        input_params: dict, optional
            If provided contains all the input parameters of the simulation.
            If it is empty than you should use this class to upload a previous experiment
        observables: TNObservables, optional
            observables used in the simulation
        """
        # Name of the input parameters in a simulation
        self._input_params_names = [
            "num_sites",
            "local_dim",
            "approach",
            "observables_filename",
            "max_bond_dimension",
            "cut_ratio",
            "trunc_tracking_mode",
            "inPATH",
            "outPATH",
            "sparse",
            "initial_state",
            "checkpoint_PATH",
            "checkpoint_frequency",
            "initial_line",
        ]

        self._input_params = {} if input_params is None else input_params
        self._from_simulation = (
            False  # flag set to True if you have to obtain data from Fortran,
        )
        # i.e. if you provide a suitable input parameter dictionary

        if all("settings%" + k in self._input_params for k in self._input_params_names):
            self.inPATH = input_params["settings%inPATH"]
            self.outPATH = input_params["settings%outPATH"]
            assert os.path.isdir(self.inPATH), "Input PATH is not a folder"
            assert os.path.isdir(self.outPATH), "Output PATH is not a folder"
            self._from_simulation = True

        self._datetime = datetime.today().strftime("%Y-%m-%d-%H:%M:%S")

        # Set to None all the variables
        self._initial_state = "Vacuum"
        if "settings%initial_state" in self._input_params:
            if self._input_params["settings%initial_state"] == "Vacuum":
                pass
            elif int(self._input_params["settings%initial_line"]) == 1:
                self._initial_state = read_mps(
                    os.path.join(
                        self.inPATH, self._input_params["settings%initial_state"]
                    )
                )
            else:
                self._initial_state = read_mps(
                    os.path.join(
                        self._input_params["settings%checkpoint_PATH"],
                        self._input_params["settings%initial_state"],
                    )
                )

        self._observables = {} if observables is None else observables
        self._singular_values_cut = None
        self._statevector = None

    def _get_results(self):
        """
        Load the results on python from the output files
        """

        assert (
            self._from_simulation
        ), "Cannot retrieve results if no input parameters are provided"

        if self._observables is not None:
            self._observables = self._observables.read(
                self._input_params["settings%observables_filename"],
                self.outPATH,
                {},
            )

            tensorlist = None
            for key in self._observables.keys():
                # The only key with the '/' is the MPS state
                if "/" in key:
                    tensorlist = read_mps(self._observables[key])
            if tensorlist is not None:
                self._observables["tn_state"] = tensorlist

        self._singular_values_cut = np.loadtxt(
            os.path.join(self.outPATH, "singular_values_cut.txt")
        )

    # ----------------------------
    # Methods to save/load results
    # ----------------------------
    def save_results(self, PATH, name, README=""):
        """Save the results in the folder PATH, creating a new subfolder named name

        Parameters
        ----------
        PATH: str
            path to the folder where to save the results
        README: str, optional
            String to save as additional README file inside the result folder.
            Default to empty string.
        name: str, optional
            Name to add before the date_time information in the folder name

        Returns
        -------
        _PATH: str
            The PATH to the new subfolder
        """
        if not os.path.isdir(PATH):
            os.makedirs(PATH)

        # Create result folder
        _PATH = os.path.join(PATH, name)
        os.makedirs(_PATH)

        # Save results
        # Input parameters
        np.save(os.path.join(_PATH, "input_params.npy"), self.input_params)
        # Singular values
        np.save(
            os.path.join(_PATH, "singular_values_cut.npy"), self.singular_values_cut
        )
        # Initial state
        np.save(os.path.join(_PATH, "initial_state.npy"), self.initial_state)
        # Observables expectation values
        np.save(os.path.join(_PATH, "observables.npy"), self.observables)
        # README
        README = self.date_time + "\n" + README
        with open(os.path.join(_PATH, "README.txt"), "w") as RM:
            RM.write(README)

        return _PATH

    def load_results(self, PATH):
        """Load the results of a previous simulation

        Parameters
        ----------
        PATH: str
            PATH to the folder from which we want to load the results

        Returns
        -------
        README: str
            The text contained in the README file inside the folder
        """
        assert os.path.isdir(PATH), f"{PATH} is not an existing folder"

        # Act on protected variables
        # Input parameters
        self._input_params = np.load(
            os.path.join(PATH, "input_params.npy"), allow_pickle="TRUE"
        ).item()
        # Singular values
        self._singular_values_cut = np.load(
            os.path.join(PATH, "singular_values_cut.npy"), allow_pickle="TRUE"
        )
        # Initial state
        self._initial_state = np.load(
            os.path.join(PATH, "initial_state.npy"), allow_pickle="TRUE"
        )
        # Probabilities
        self._observables = np.load(
            os.path.join(PATH, "observables.npy"), allow_pickle="TRUE"
        ).item()
        # README
        with open(os.path.join(PATH, "README.txt"), "r") as RM:
            README = RM.read()
        self._datetime = README.split("\n")[0]

        return README

    def set_results(self, result_dict, singvals_cut):
        """Set the results of a simulation

        Parameters
        ----------
        result_dict: dict
            Dictionary of the attribute to be set
        singvals_cut: array-like
            Array of singular values cut through the simulation
        """
        self._observables = result_dict
        self._singular_values_cut = singvals_cut

        tensorlist = None
        for key in result_dict.keys():
            # The only key with the '/' is the MPS state
            if "/" in key:
                tensorlist = read_mps(result_dict[key])
        if tensorlist is not None:
            self._observables["tn_state"] = tensorlist

    # -----------------------------
    # Methods to access the results
    # -----------------------------
    @property
    def fidelity(self):
        """
        Return the lower bound for the fidelity of the simulation,
        using the method described in
        If you are interested in the evolution of the fidelity through
        the simulation compute it yourself using `np.cumprod(1-self.singular_values_cut)`.

        Returns
        -------
        float
            fidelity of the final state
        """

        fid = np.prod(1 - np.array(self.singular_values_cut))
        return fid

    @property
    def measures(self):
        """Obtain the measures of the simulation as a dictionary.
        The keys are the measured states, the values the number of occurrencies

        Returns
        -------
        measures: dict
            Measures of the simulation
        """
        if "projective_measurements" in self.observables.keys():
            measures = self.observables["projective_measurements"]
        else:
            measures = None
        return measures

    @property
    def statevector(self):
        """Obtain the statevector as a complex numpy array

        Returns
        -------
        statevector: np.array or None
            The statevector of the simulation
        """
        if self._statevector is None:
            if "tn_state" in self.observables.keys():
                tensor_list = self.observables["tn_state"]
                if isinstance(tensor_list[0], list):
                    tn_state = TTN.from_tensor_list(tensor_list)
                else:
                    tn_state = MPS.from_tensor_list(tensor_list)
                if tn_state.num_sites < 30:
                    self._statevector = tn_state.to_statevector(qiskit_order=True)
        return self._statevector

    @property
    def singular_values_cut(self):
        """Obtain the singular values cutted through the simulation, depending on the mode
        chosen. If 'M' for maximum (default), 'C' for cumulated.

        Returns
        -------
        singular_values_cut: np.array
            Singular values cut during the simulation
        """
        return self._singular_values_cut

    @property
    def computational_time(self):
        """Obtain the computational time of the simulation

        Returns
        -------
        computational_time: double
            computational time of the simulation
        """
        if "time" in self.observables.keys():
            time = self.observables["time"]
        else:
            time = None
        return time

    @property
    def entanglement(self):
        """Obtain the bond entanglement entropy measured along each bond of the MPS at
        the end of the simulation

        Returns
        -------
        entanglement: dict or None
            Bond entanglement entropy
        """
        if "bond_entropy" in self.observables.keys():
            entanglement = self.observables["bond_entropy"]
        elif "bond_entropy0" in self.observables.keys():
            entanglement = self.observables["bond_entropy0"]
        else:
            entanglement = None
        return entanglement

    @property
    def measure_probabilities(self):
        """Return the probability of measuring a given state, which is computed using a
        binary tree by eliminating all the branches with probability under a certain threshold.

        Returns
        -------
        measure_probabilities: dict or None
            probability of measuring a certain state if it is greater than a threshold
        """
        keys = ["unbiased_probability", "even_probability", "greedy_probability"]
        probs = []
        for key in keys:
            if key in self.observables.keys():
                probs += [self.observables[key]]
            else:
                probs += [None]
        return probs

    @property
    def date_time(self):
        """Obtain the starting date and time of the simulation, in the format
        ``Year-month-day-Hour:Minute:Second``

        Returns
        -------
        datetime: string
            The date-time when the simulation started
        """
        return self._datetime

    @property
    def input_params(self):
        """Obtain the input parameters used in the simulation, which are the following:
            - 'num_sites',              number of sites of the mps
            - 'local_dim',              local dimension of the single site
            - 'max_bond_dim',           maximum bond dimension of the mps
            - 'cut_ratio',              cut ration used in the SVD truncation
            - 'in_name',                path to the fortran input folder
            - 'out_name',               path to the Fortran output folder
            - 'trunc_tracking_mode',           mode to save the singular values cut
            - 'sparse',                 if the input gate tensors are saved as sparse
            - 'par_approach',           parallel approach of the simulation. Can be
                'SR' (serial), 'MW' (master/workers) or 'CT' (cartesian)
            - 'initial_state',          initial state of the simulation. 'Vacuum' or
                the PATH to the initial state
            - 'do_observables',         True if there are some observables to compute at
                the end

        Returns
        -------
        input_params: dict or None
            Input parameters
        """
        return self._input_params

    @property
    def tens_net(self):
        """
        Returns the tensor list in row-major format.
        The indexing of the single tensor is as follows:

        .. code-block::

            1-o-3
             2|

        Returns
        -------
        mps: list
            list of np.array tensors
        """
        if "tn_state" in self.observables.keys():
            tn_state = self.observables["tn_state"]
        else:
            tn_state = None
        return tn_state

    @property
    def initial_state(self):
        """Returns the initial state of the simulation, as an MPS in row-major format or as
        a string if starting from the Vacuum state

        Returns
        -------
        initial_state: list or str
            list of np.array tensors or Vacuum
        """
        return self._initial_state

    @property
    def observables(self):
        """Returns the expectation values of the observables as a dict with the format
            observable_name : observable_expectation_value

        Returns
        -------
        observables: dict or None
            Expectation values of the observables
        """
        return self._observables


class QCCheckpoints:
    """
    Class to handle checkpoint parameters

    Parameters
    ----------
    PATH: str, optional
        PATH to the checkpoint directory. Default `data/checkpoints/`.
    frequency: float, optional
        Decide the frequency, in **hours**, of the checkpoints.
        If negative no checkpoints are present. Default to -1.
    input_nml: str, optional
        Name of the input namelist. Default 'input.nml'
    """

    def __init__(self, PATH="data/checkoints/", frequency=-1, input_nml="input.nml"):
        self._PATH = PATH if (PATH.endswith("/")) else PATH + "/"
        self._frequency = frequency
        self._input_nml = input_nml

    def set_up(
        self,
        input_dict,
        operators=QCOperators(),
        observables=TNObservables(),
        circ_str="",
    ):
        """Set up the checkpoints directory

        Parameters
        ----------
        input_dict : dict
            Input parameter dictionary
        operators : :py:class:`QCOperators`, optional
            Tensor operators
        obervables : :py:class: `TNObservables`, optional
            Tensor observables
        circ_str: str
            circuit string
        """
        if not isinstance(operators, QCOperators):
            raise TypeError("Operators must be QCOperators type")
        elif not isinstance(observables, TNObservables):
            raise TypeError("observables must be TNObservables type")

        if not os.path.isdir(self.PATH):
            os.mkdir(self.PATH)

        # Modify for new PATH
        input_dict["inPATH"] = self.PATH

        # Write files that can be already written
        with open(os.path.join(self.PATH, "observables.pk"), "wb") as fh:
            pickle.dump(observables, fh)
        _, operator_mapping = operators.write_input_3(self.PATH)
        observables.write(self.PATH, {}, operator_mapping)
        write_nml("INPUT", input_dict, os.path.join(self.PATH, self.input_nml))
        with open(os.path.join(self.PATH, "circuit.dat"), "w") as fh:
            fh.write(circ_str)

    @property
    def PATH(self):
        """PATH property"""
        return self._PATH

    @property
    def frequency(self):
        """Checkpoint frequency property"""
        return self._frequency

    @property
    def input_nml(self):
        """Input namelist property"""
        return self._input_nml

    def to_dict(self):
        """Return the ordered dictionary of the properties of
        the class

        Returns
        -------
        dictionary: OrderedDict
            Ordered dictionary of the class properties
        """
        dictionary = OrderedDict()

        dictionary["checkpoint_PATH"] = self.PATH
        dictionary["checkpoint_frequency"] = self.frequency
        dictionary["initial_line"] = 1

        return dictionary


class QCIO:
    """
    Class to handle Input/Output parameters

    Parameters
    ----------
    inPATH: str, optional
        PATH to the directory containing the input files.
        Default to 'data/in/'
    outPATH: str, optional
        PATH to the directory containing the output files.
        Default to 'data/out/'
    input_namelist: str, optional
        Name of the input namelist file. Name, NOT PATH.
        Default to 'input.nml'
    exe_file: list of str, optional
        Path to the executable plus additional commands
        Default to `[EXE_PATH_SERIAL]`.
    initial_state: str or :py:class:`MPS`, optional
        If an MPS, then the list of tensors is used as initial state for
        a starting point of the FORTRAN simulation, saving the file to
        inPATH/initial_state.dat. If 'Vacuum' start from |000...0>. Default to 'Vacuum'.
        If a PATH it is a PATH to a saved MPS.
    sparse: bool, optional
        Weather to write operators in a semi-sparse format or not.
        Default to False
    """

    def __init__(
        self,
        inPATH="data/in/",
        outPATH="data/out/",
        input_namelist="input.nml",
        exe_file=None,
        initial_state="Vacuum",
        sparse=False,
    ):
        self._inPATH = inPATH if inPATH.endswith("/") else inPATH + "/"
        self._outPATH = outPATH if outPATH.endswith("/") else outPATH + "/"
        self._input_namelist = input_namelist

        if exe_file is None:
            self._exe_file = [EXE_PATH]
        # Check if people don't read docstring
        # and use strings nevertheless
        elif isinstance(exe_file, str):
            self._exe_file = [exe_file]
        else:
            self._exe_file = exe_file
        self._sparse = sparse
        self._initial_state = initial_state

    def setup(self):
        """Setup the io files"""

        # Directories
        if not os.path.isdir(self.inPATH):
            os.makedirs(self.inPATH)
        if not os.path.isdir(self.outPATH):
            os.makedirs(self.outPATH)

        # Executable
        if self.exe_file[-1] != os.path.join(self.inPATH, self.input_namelist):
            self._exe_file += [os.path.join(self.inPATH, self.input_namelist)]

        # Initial state
        if self.initial_state != "Vacuum":
            if isinstance(self.initial_state, str):
                # Handle the string case assuming it is a path
                if not os.path.isfile(self.initial_state):
                    raise Exception("Path to input file does not exist.")
                else:
                    new_path = os.path.join(
                        self.inPATH, os.path.basename(self.initial_state)
                    )
                    if not os.path.samefile(self._initial_state, new_path):
                        shutil.move(self.initial_state, new_path)
                        self._initial_state = os.path.basename(self.initial_state)
            else:
                # Assume it is an MPS that we can write
                self.initial_state.write(os.path.join(self.inPATH, "initial_state.dat"))
                self._initial_state = "initial_state.dat"

    @property
    def inPATH(self):
        """Input PATH property"""
        return self._inPATH

    @property
    def exe_cmd(self):
        """Executable command to run on the terminal"""
        return [self.exe_file, self._inPATH + self._input_namelist]

    @property
    def outPATH(self):
        """Output PATH property"""
        return self._outPATH

    @property
    def input_namelist(self):
        """Input namelist property"""
        return self._input_namelist

    @property
    def exe_file(self):
        """Executable file and commands property"""
        return self._exe_file

    @property
    def sparse(self):
        """Tensor sparsity property"""
        return self._sparse

    @property
    def initial_state(self):
        """Initial state property"""
        return self._initial_state

    # @initial_state.setter
    def set_initial_state(self, initial_state):
        """Modify the initial state property"""
        if not isinstance(initial_state, str):
            if not isinstance(initial_state, MPS):
                raise TypeError(
                    "A non-str initial state must be initialized as MPS class"
                )
        self._initial_state = initial_state

    # @exe_file setter
    def set_exe_file(self, exe_file):
        """Modify exe file"""
        if not isinstance(exe_file, list):
            raise TypeError(f"exe_file must be a list of strings, not {type(exe_file)}")
        self._exe_file = exe_file

    def to_dict(self):
        """Return the ordered dictionary of the properties of
        the class

        Returns
        -------
        dictionary: OrderedDict
            Ordered dictionary of the class properties
        """
        dictionary = OrderedDict()
        for prop, value in vars(self).items():
            if prop in ("_exe_file", "_input_namelist"):
                continue
            dictionary[prop[1:]] = value

        return dictionary


class QCConvergenceParameters(TNConvergenceParameters):
    """Convergence parameter class, inhereting from the
    more general Tensor Network type. Here the convergence
    parameters are only the bond dimension and the cut ratio.

    Parameters
    ----------
    max_bond_dimension : int, optional
        Maximum bond dimension of the problem. Default to 10.
    cut_ratio : float, optional
        Cut ratio for singular values. If :math:`\\lambda_n/\\lambda_1 <` cut_ratio then
        :math:`\\lambda_n` is neglected. Default to 1e-9.
    trunc_tracking_mode : str, optional
        Modus for storing truncation, 'M' for maximum, 'C' for
        cumulated (default).

    """

    def __init__(
        self,
        max_bond_dimension=10,
        cut_ratio=1e-9,
        trunc_tracking_mode="C",
        singval_mode=None,
    ):
        if singval_mode is not None:
            warn(
                "singval_mode parameter is deprecated. Please use trunc_tracking_mode instead"
            )
            trunc_tracking_mode = singval_mode
        TNConvergenceParameters.__init__(
            self,
            max_bond_dimension=max_bond_dimension,
            cut_ratio=cut_ratio,
            trunc_tracking_mode=trunc_tracking_mode,
        )

    def to_dict(self):
        """Return the ordered dictionary of the properties of
        the class

        Returns
        -------
        dictionary: OrderedDict
            Ordered dictionary of the class properties
        """
        dictionary = OrderedDict()
        dictionary["max_bond_dimension"] = self.max_bond_dimension
        dictionary["cut_ratio"] = self.cut_ratio
        dictionary["trunc_tracking_mode"] = self.trunc_tracking_mode

        return dictionary


class QCBackend:
    """
    Backend for the simulation. Contains all the informations about
    which executable you want to run

    Parameters
    ----------
    backend : str, optional
        First backend definition. Either "PY" (python) or "FR" (fortran).
        Default to "PY".
    precision: str, optional
        Precision of the simulation. Either "Z" (double complex) or "C"
        (single complex). Default to "Z".
    device: str, optional
        Device of the simulation. Either "cpu" or "gpu". Default to "cpu".
    num_procs: int, optional
        Number of processes for the MPI simulation. Default to 1.
    mpi_approach: str, optional
        Approach for the MPI simulation. Either "MW", "CT" or "SR".
        Default to "SR".
    ansatz : str, optional
        Weather to run the circuit with MPS or TTN tensor network ansatz.
        Default to "MPS".
    """

    def __init__(
        self,
        backend="PY",
        precision="Z",
        device="cpu",
        num_procs=1,
        mpi_approach="SR",
        ansatz="MPS",
    ):
        if backend == "PY" and mpi_approach != "SR":
            raise ValueError("Only serial simulation available in python")
        if backend == "FR" and device == "gpu" and precision == "C":
            raise ValueError(
                "Only double precision complex available "
                + "in fortran with the GPU device"
            )
        if num_procs == 1:
            mpi_approach = "SR"

        self._backend = backend.upper()
        self._precision = precision.upper()
        self._device = device
        self._num_procs = num_procs
        self._mpi_approach = mpi_approach.upper()
        if backend == "FR" and ansatz.upper() != "MPS":
            warn(
                f"Only MPS ansatz available on fortran simulation, not {ansatz}."
                + "The ansatz is set back to MPS."
            )
            ansatz = "MPS"
        self._ansatz = ansatz.upper()

    def to_dict(self):
        """
        Map the backend to a dictionary for fortran
        """
        dictionary = OrderedDict({})
        mpi = "T" if self._num_procs > 1 else "F"
        gpu = "T" if self.device == "gpu" else "F"
        dictionary["simulation_mode"] = self.precision + mpi + gpu
        dictionary["approach"] = self.mpi_approach
        # dictionary["ansatz"] = self.ansatz

        return dictionary

    @property
    def backend(self):
        """Backend property"""
        return self._backend

    @property
    def precision(self):
        """Precision property"""
        return self._precision

    @property
    def device(self):
        """Device property"""
        return self._device

    @property
    def num_procs(self):
        """Number of processes property"""
        return self._num_procs

    @property
    def mpi_approach(self):
        """mpi_approach property"""
        return self._mpi_approach

    @property
    def ansatz(self):
        """ansatz property"""
        return self._ansatz

    @property
    def identifier(self):
        """Identifier combining all properties."""
        return ":".join(
            [
                self.backend,
                self.precision,
                self.device,
                str(self.num_procs),
                self.mpi_approach,
                self.ansatz,
            ]
        )


def merge_ordered_dicts(dicts):
    """Merge ordered dicts together, concatenating them in the order provided in the list

    Parameters
    ----------
    dicts : list of OrderedDict
        OrderedDict to concatenate

    Return
    ------
    final_dict: OrderedDict
        Concatenated OrderedDict
    """
    for dictionary in dicts:
        if not isinstance(dictionary, OrderedDict):
            raise TypeError("Only OrderedDict can be concatenated using this function")

    final_dict = dicts[0]
    for dictionary in dicts[1:]:
        final_dict.update(dictionary)

    return final_dict


def print_state(dense_state):
    """
    Prints a *dense_state* with kets. Compatible with quimb states.

    Parameters
    ----------
    dense_state: array_like
            Dense representation of a quantum state

    Returns
    -------
    None: None
    """

    NN = int(np.log2(len(dense_state)))

    binaries = [bin(ii)[2:] for ii in range(2**NN)]
    binaries = ["0" * (NN - len(a)) + a for a in binaries]  # Pad with 0s

    ket = []
    for ii, coef in enumerate(dense_state):
        if not np.isclose(np.abs(coef), 0.0):
            if np.isclose(np.imag(coef), 0.0):
                if np.isclose(np.real(coef), 1.0):
                    ket.append("|{}>".format(binaries[ii]))
                else:
                    ket.append("{:.3f}|{}>".format(np.real(coef), binaries[ii]))
            else:
                ket.append("{:.3f}|{}>".format(coef, binaries[ii]))
    print(" + ".join(ket))


def fidelity(psi, phi):
    """
    Returns the fidelity bewteen two quantum states *psi*, *phi* defined as
    :math:`|\\langle\\psi|phi\\rangle|^2`

    Parameters
    ----------
    psi: complex np.array or quimb.core.qarray
            Quantum state
    phi: complex np.array or quimb.core.qarray
            Quantum state

    Returns
    -------
    Fidelity: double real
            Fidelity of the two quantum states
    """

    # Enforcing normalization
    psi /= np.sqrt(np.abs(np.sum(psi.conj() * psi)))
    phi /= np.sqrt(np.abs(np.sum(phi.conj() * phi)))

    return np.abs(np.vdot(psi, phi)) ** 2



import os
import os.path
import unittest
from io import StringIO
import pylint.lint
from pylint.reporters.text import ParseableTextReporter


class TestPylint(unittest.TestCase):
    """
    Run pylint to check syntax in source files.

    **Details**

    We disable globally:

    * C0325: superfluous parenthesis
    * C0209: consider using fstring
    * W1514: unspecified encoding
    * R1711: useless returns (for allowing empty iterators with
      return-yield)
    * R1732: consider using with
    * Skip Unused argument errors when args
    * Skip Unused argument errors when kargs

    """

    def setUp(self):
        """
        Provide the test setup.
        """
        pattern_1 = (
            "[E0611(no-name-in-module), ] No name '_TNObsBase' in module"
            + " 'qtealeaves'"
        )
        pattern_2 = "[E0401(import-error), ] Unable to import 'strawberryfields'"
        pattern_3 = "[E0401(import-error), ] Unable to import 'strawberryfields.backends.fockbackend'"
        self.default_patterns = [pattern_1, pattern_2, pattern_3]
        self.pylint_args = {
            "good-names": "ii,jj,kk,nn,mm,fh,dx,dy,dz,dt,hh,qc,dl,gh,xp",
            "disable": "C0325,C0209,W1514,R1711,R1732"
            #'ignore_in_line' : pattern_1
        }

    def run_pylint(self, filename, local_settings={}):
        """
        Run linter test with our unit test settings for one specific
        filename.
        """
        args = []

        ignore_in_line = self.default_patterns
        if "ignore_in_line" in local_settings:
            ignore_in_line += local_settings["ignore_in_line"]
            del local_settings["ignore_in_line"]

        for elem in self.pylint_args.keys():
            args += ["--" + elem + "=" + self.pylint_args[elem]]

            if elem in local_settings:
                args[-1] = args[-1] + "," + local_settings[elem]
                del local_settings[elem]

        for elem in local_settings.keys():
            args += ["--" + elem + "=" + local_settings[elem]]

        args += [filename]

        obj = StringIO()
        reporter = pylint.reporters.text.ParseableTextReporter(obj)
        pylint.lint.Run(args, reporter=reporter, exit=False)

        error_list = []
        for elem in obj.getvalue().split("\n"):
            tmp = elem.replace("\n", "")

            if len(tmp) == 0:
                continue
            if tmp.startswith("***"):
                continue
            if tmp.startswith("---"):
                continue
            if tmp.startswith("Your code"):
                continue
            if "Unused argument 'args'" in tmp:
                continue
            if "Unused argument 'kwargs'" in tmp:
                continue

            do_continue = False
            for pattern in ignore_in_line:
                if pattern in tmp:
                    do_continue = True

            if do_continue:
                continue

            error_list.append(tmp)

        return error_list

    def test_folders_recursively(self):
        """
        Recursively run python linter test on all .py files of
        specified folders.
        """
        parent_folders = ["qmatchatea"]
        skip_files = []
        error_list = []

        for elem in parent_folders:
            for root, dirnames, filenames in os.walk(elem):
                for filename in filenames:
                    if not filename.endswith(".py"):
                        continue

                    if filename in skip_files:
                        continue

                    target_file = os.path.join(root, filename)

                    target_attr = "get_settings_" + filename.replace(".py", "")
                    if hasattr(self, target_attr):
                        target_setting = self.__getattribute__(target_attr)()
                    else:
                        target_setting = {}

                    error_list_ii = self.run_pylint(
                        target_file, local_settings=target_setting
                    )

                    error_list += error_list_ii

        if len(error_list) > 0:
            print("\n".join(error_list))

        self.assertEqual(len(error_list), 0, "\n".join(error_list))

    # --------------------------------------------------------------------------
    #                          Settings for qtealeaves
    # --------------------------------------------------------------------------

    @staticmethod
    def get_settings_mps_simulator():
        """
        Linter settings for mps simulator.

        **Details**

        We locally ignore:

        * R0913: too many arguments
        * R0914: too many locals
        * R1720: no-else-raise
        """
        return {
            "ignore_in_line": ["R0913", "R0914", "R1720"],
            "good-names": "ii, jj, kk, ss, fh, tSVD, UU, Vh, QQ, RR, op, xp",
        }

    @staticmethod
    def get_settings_sf_utils():
        """
        Linter settings for mps simulator.

        **Details**

        We locally ignore:

        * W0223: abstract-method
        * R0903: too few public methods
        * C0103: invalid name
        * W0106: expression not assigned
        """
        return {"disable": "W0223, R0903, C0103, W0106"}

    @staticmethod
    def get_settings_mpi_utils():
        """
        Linter settings for mps simulator.

        **Details**

        We locally ignore:

        * C0301: line-too-long
        """
        return {"disable": "C0301"}

    @staticmethod
    def get_settings_qk_utils():
        """
        Linter settings for mps simulator.

        **Details**

        We locally ignore:

        * C0103: invalid name
        * R1705: no-else-return
        """
        return {"disable": "C0103, R1705"}

    @staticmethod
    def get_settings_tn_utils():
        """
        Linter settings for mps simulator.

        **Details**

        We locally ignore:

        * R1732: consider-using-with
        * W0237: argument renamed
        * W0221: argument different
        """
        return {"disable": "R1732, W0237,W0221"}

    @staticmethod
    def get_settings_utils():
        """
        Linter settings for mps simulator.

        **Details**

        We locally ignore:

        * C0103: invalid name
        * C0302: too-many-lines
        * R1720: no-else-raise
        * R0902: too-many-instance-attributes
        * R0913: too-many-arguments
        """
        return {"disable": "C0103, C0302, R1720, R0902, R0913"}

    @staticmethod
    def get_settings_preprocessing():
        """
        Linter settings for mps simulator.

        **Details**

        We locally ignore:

        * R0914: too-many-locals
        """
        return {"disable": "R0914"}

    @staticmethod
    def get_settings_par_simulations():
        """
        Linter settings for mps simulator.

        **Details**

        We locally ignore:

        * C0103: invalid name
        * R1720: no-else-raise
        * E1120: no-value-parameter
        * R1705: no-else-return
        """
        return {"disable": "R1720, C0103, E1120, R1705"}

    @staticmethod
    def get_settings_py_emulator():
        """
        Linter settings for mps simulator.

        **Details**

        We locally ignore:

        * R1720: no-else-raise
        * R1724: no-else-continue
        * R0913: too-many-arguments
        * R0914: too-many-locals
        * R0912: too-many-branches
        * R0915: too-many-statements
        * E1121: too-many-function-args,
        """
        return {"disable": "R1720, R1724, R0913, R0912, R0914, R0915, E1121"}

    @staticmethod
    def get_settings_interface():
        """
        Linter settings for mps simulator.

        **Details**

        We locally ignore:


        * R1720: no-else-raise
        * R0913: too-many-arguments
        * W1510: subprocess-run-check
        * W0212: protected-access
        * R0912: too-many-branches
        * R0915: too-many-statements
        * R0914: too many locals
        * R0902: too-maby-instance-attributs
        * C0103: invalid name


        """
        return {"disable": "R1720,R0913, W1510,W0212,R0915,R0912,C0103,R0914,R0902"}

    @staticmethod
    def get_settings_circuit():
        """
        Linter settings for local observables.

        **Details**

        We locally ignore:

        C0103 : invalid-name
        C0302 : too-many-lines
        R1733 : unnecessary-dict-index-lookup
        R0913 : too-many-arguments
        R0904 : too-many-public-methods
        R0902 : too-maby-instance-attributes

        """
        pattern = (
            "[R0912(too-many-branches), Qcircuit.to_matrix] Too many branches (17/12)"
        )
        return {
            "disable": "C0103, C0302, R1733, R0913, R0902, R0904",
            "ignore_in_line": [pattern],
        }

    @staticmethod
    def get_settings___init__():
        """
        Linter settings for mps simulator.

        **Details**

        We locally ignore:


        * E0602 : undefined variable


        """
        return {"disable": "E0602"}



# This code is part of qmatchatea.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""
I/O functions for tensors and mps (list of tensors).
Furthermore, the class :py:class:`QCOperators` for handling operators is provided.

Functions and classes
---------------------

"""

import os
import os.path
from ast import literal_eval
from typing import OrderedDict
import numpy as np
from qtealeaves import read_tensor
from qtealeaves.operators import TNOperators
from qtealeaves.fortran_interfaces import StrBuffer

__all__ = ["write_tensor", "write_mps", "read_mps", "QCOperators"]

pauli_matrices = {
    "Z": np.array([[1, 0], [0, -1]]),
    "X": np.array([[0, 1], [1, 0]]),
    "Y": np.array([[0, -1j], [1j, 0]]),
    "I": np.array([[1, 0], [0, 1]]),
}


def read_nml(namelist_file):
    """
    Read a fortran namelist file

    Parameters
    ----------
    namelist_file : str
        PATH to the namelist

    Returns
    -------
    namelist_dict : OrderedDict
        Dictionary to be eventually passed to write_nml
    """
    if not os.path.isfile(namelist_file):
        raise IOError(f"{namelist_file} is not a file")
    with open(namelist_file, "r") as fh:
        file = fh.readlines()
    namelist_name = file[0].replace("&", "").replace("\n", "")

    namelist_dict = OrderedDict()
    for line in file[1:-2]:
        line = line.replace(" ", "").replace("\n", "")
        data = line.split("=")
        if data[1] == ".true.":
            namelist_dict[data[0]] = True
        elif data[1] == ".false.":
            namelist_dict[data[0]] = False
        else:
            namelist_dict[data[0]] = literal_eval(data[1])

    return namelist_name, namelist_dict


def write_tensor(tensor, dest, cmplx=True, sparse=False):
    """
    Write a tensor stored in a numpy matrix to a file. Conversion
    to column major is taken care of here.

    **Arguments**

    tensor : np.ndarray
        Tensor to be written to the file.

    dest : str, or filehandle
        If string, file will be created or overwritten. If filehandle,
        then data will be written there.

    sparse: bool
        If True write the tensor in a sparse format, i.e. each row is written as
        idx elem
        where idx is the position of the element elem in the tensor vector
    """
    if isinstance(dest, str):
        fh = open(dest, "w+")
    elif hasattr(dest, "write"):
        fh = dest

    # Number of links
    fh.write("%d \n" % (len(tensor.shape)))

    # Dimensions of links
    dl = " ".join(list(map(str, tensor.shape)))
    fh.write(dl + "\n")

    # Now we need to transpose
    tensor_colmajor = np.reshape(tensor, -1, order="F")
    if sparse:
        nonzero = np.nonzero(tensor_colmajor)[0]  # index of nonzero element
        tensor_colmajor = tensor_colmajor[nonzero]
        nonzero += 1  # from python to fortran indexing
        fh.write(f"{len(nonzero)} \n")

    for ii, elem in enumerate(tensor_colmajor):
        if cmplx:
            if sparse:
                fh.write(
                    "%d (%30.15E, %30.15E)\n"
                    % (nonzero[ii], np.real(elem), np.imag(elem))
                )
            else:
                fh.write("(%30.15E, %30.15E)\n" % (np.real(elem), np.imag(elem)))
        else:
            if sparse:
                fh.write("%d %30.15E\n" % (nonzero[ii], np.real(elem)))
            else:
                fh.write("%30.15E\n" % (np.real(elem)))
            imag_part = np.imag(elem)
            if np.abs(imag_part) > 1e-14:
                raise Exception("Writing complex valued tensor as real valued tensor.")

    if isinstance(dest, str):
        fh.close()

    return


def read_mps(filename, cmplx=True):
    """Read an MPS written by FORTRAN in a formatted way on file.
    Reads in column-major order but the output is in row-major

    Parameters
    ----------
    filename: str
            PATH to the file
    cmplx: bool, optional
            If True the MPS is complex, real otherwise. Default to True

    Returns
    -------
    mps: list
            list of np.array in row-major order
    """
    mps = []
    with open(filename, "r") as fh:
        num_sites = int(fh.readline())
        for _ in range(num_sites):
            tens = read_tensor(fh, cmplx=cmplx)
            mps.append(tens)

    return mps


def write_mps(filename, mps, cmplx=True):
    """Write an MPS in python format into a FORTRAN format, i.e.
    transforms row-major into column-major

    Parameters
    ----------
    filename: str
            PATH to the file
    mps: list
            List of tensors forming the mps
    cmplx: bool, optional
            If True the MPS is complex, real otherwise. Default to True

    Returns
    -------
    None
    """
    with open(filename, "w") as fh:
        fh.write(str(len(mps)) + " \n")
        for tens in mps:
            write_tensor(tens, fh, cmplx=cmplx)

    return None


class QCOperators(TNOperators):
    """
    Class to store and save to file operators. To add an operator to the list use
    self.ops['op_name'] = op_tensor.

    It starts with the pauli operators already defined, with keywords "X","Y","Z"
    """

    def __init__(self, folder_operators="TENSORS"):
        TNOperators.__init__(self, folder_operators=folder_operators)
        for key, val in pauli_matrices.items():
            self.ops[key] = val

    def write_operator(self, folder_dst, operator_name, params, sparse=False):
        """
        Write operator to file.

        Parameters
        ---------

        folder_dst : str or filehandle
            If filehandle, write there. If string, it defines the folder.
        operator_name : str
            Name of the operator.
        params : dictionary
                Contains the simulation parameters.
        kwargs : passed to write_symtensor

        """
        if operator_name not in self.ops:
            raise Exception("Operator `%s` not defined." % (operator_name))
        if hasattr(folder_dst, "write"):
            # filehandle
            full_filename = folder_dst
        else:
            full_filename = os.path.join(folder_dst, operator_name + ".dat")

        operator = self.get_operator(operator_name, params)

        write_tensor(operator, full_filename, sparse=sparse)

        return operator_name + ".dat"

        # tensor_backend is put to 2 since we only use non-symmetric tensors
        # Temporarily disabled quick inhereted way because of sparse tensors
        # return TNOperators.write_operator(self, folder_dst, operator_name,
        # params, tensor_backend=2, sparse=sparse)

    def write_input_3(self, folder_name, filename="operators.dat", sparse=False):
        """Write on file for fortran the operators used in the observable

        Parameters
        ----------
        folder_name: str
            PATH to the folder where we want to save the file
        filename: str, optional
            Name of the file. Default to 'operators.dat'
        sparse: bool, optional
            If True write the operators in a semi-sparse format, i.e. by writing
            only the non-zero elemnts and their position. Default to False.

        Returns
        -------
        relative_file: str
            Full PATH to the observable file 'observables.dat'
        operator_id_mapping: dict
            Dictionary which maps the operators names to their IDs
        """

        full_path = os.path.join(folder_name, self.folder_operators)
        if not os.path.isdir(full_path):
            os.makedirs(full_path)

        relative_file = os.path.join(full_path, filename)
        buffer_str = StrBuffer()

        operator_id_mapping = {}

        ii = 0
        for operator_ii in self.ops.keys():
            ii += 1
            # Write the operator on the buffer string
            _ = self.write_operator(buffer_str, operator_ii, None, sparse=sparse)
            # Mapping between the operator name and a number
            operator_id_mapping[operator_ii] = ii

        # Apriori length because operators are not written
        nn = len(operator_id_mapping)

        fh = open(relative_file, "w+")
        fh.write(str(nn) + "\n")
        fh.write(buffer_str())
        fh.close()

        return relative_file, operator_id_mapping



# This code is part of qmatchatea.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""
Core functions to run the Fortran MPS simulation. In particular, the :py:func:`run_simulation`
function is the most important one, and one should make particular attention to it.

Running the program
-------------------

 It is possible to choose between different approaches for running the simulation:
 - The backend can be either python "PY" or fortran "FR".
 - The machine precision can be either double complex "Z" or single complex "C".
 - The device of the simulation can be either "cpu" or "gpu".
 - The number of processes for the MPI simulation. If 1 is passed, then a serial
   program is run.
 - The approach for the MPI simulation. Either master/worker "MW", cartesian "CT" or
   serial "SR".


Serial program
~~~~~~~~~~~~~~

The default way of running this program is to run it serially. While it may be slower than a
parallel implementation it is safer, since the approximation we do on the state is minor.
There are different possibilities to run the program serially:

- Using the Fortran backend (``mpi_approach='SR'``). It is the faster implementation, but requires
  the presence of the executable ``main_qmatchatea.exe``.
- Using the Python backend (``backend='PY'``). The backend is entirely built with numpy, it
  is optimized but slightly slower than the fortran backend.
- Using the Python GPU backend (``backend='PY'``, ``device='gpu'``).
  The backend is entirely built with cupy,
  and runs the simulation on a GPU, giving back the results on the CPU.
- Using the Python Fortran backend (``approach='PF'``). This backend uses the Python backend
  but simulates the same I/O files of the fortran one. It can be used as an initial debugging
  backend in the case the I/O fortran files are used in a more general setting. Furthermore,
  the strawberry-fields interface is not available for the ``'PY'`` approach, but is available
  here. Finally, it is very useful for developers.

Parallel program
~~~~~~~~~~~~~~~~

There are three possible algorithm provided to run the program on multiple cores:

- a Master/Worker approach (``mpi_approach='MW'``), where you store the state in the master
  process and perform the application of the evolution operators on the workers.
  It has the advantages that fewer workers are idle during the computation.
- a Cartesian approach (``mpi_approach='CT'``), where the MPS is evenly divided among different
  processes, and each process perform the evolution only on his subsystem
- running multiple independent serial simulations on multiple cores. To know more see
  :doc:`par_simulations`

At the end of the evolution the MPS is brought back together into the master process and the
measurements are performed. The command to call the parallel program should be
:code:`mpirun -np np ./par_main_qmatchatea.exe` where :code:`np` is the
number of processes you will use. It should be given as input to the :class:`IO_INFO`.
For further details on the input class see :doc:`utils`.

Remember that there are communications involved in the MPI parallel program.
You should use it only if your system is large enough or if the bond
dimension used is big enough.

Changing the approach at runtime with command line arguments
~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

When you write a python script with a specific approach, it is possible to change that
approach at runtime, using optional arguments. The following optional arguments are
available:

- ``--approach``: it should be a ``str``, and if it is present, it overwrites the script approach.
  This means that, if in your ``my_script.py`` you selected the approach ``PF`` but you run on the
  command line ``python my_script.py --approach='SR'``, the :func:`run_simulation` will use the
  approach ``'SR'``.
- ``--mpi``: it should be an ``int``. It represent the number of cores that should be used in the
  simulation. If ``--approach`` is absent, it overwrites the script approach to ``'CT'``, and then
  sets the number of cores to the numeric value passed. It automatically changes the PATH of the
  executable to the parallel path.

Checking the convergence: Singular values cut
---------------------------------------------

The singular values cut are an indication of the approximation taking place during the
simulation. Indeed, at each application of two-qubit there is an approximation going on,
subject to the *bond_dim* :math:`\\chi` and the *cut_ratio* :math:`\\epsilon` parameters.
Based on the *trunc_tracking_mode* the singular values are reported:

- *trunc_tracking_mode=M*, at each two-site operator we save the maximum of the singular values cut
- *trunc_tracking_mode=C*, at each two-site operator we save the **sum** of the singular values cut

We recall that, given a decreasing-ordered vector of singular values
:math:`S=(s_1, s_2, \\dots, s_n)` we truncate all the singular values:

.. math::

        s_i \\; \\mbox{ is truncated }\\quad \\mbox{if} \\; i>\\chi \\; \\mbox{ or } \\;
        \\frac{s_i}{s_1}<\\epsilon

Observables
-----------

For a detailed description of the observables, i.e. the measurable quantities, please refer
to :doc:`observables`

"""

# Numpy
import os
import os.path
from typing import OrderedDict
import subprocess
import sys
import pickle
import argparse

import numpy as np
from qtealeaves import write_nml
from qtealeaves.observables import TNObservables

from qmatchatea.py_emulator import mock_fortran, run_py_simulation
from .preprocessing import pre_parser, preprocess
from .tn_utils import QCOperators, read_nml
from .mpi_utils import to_layered_circ
from .qk_utils import qk_transpilation_params
from .utils import QCCheckpoints, QCIO, QCConvergenceParameters, merge_ordered_dicts
from .utils import simulation_results, EXE_PATH, QCBackend

__all__ = ["run_simulation", "simulation_results", "restart_from_checkpoint"]


def run_simulation(
    circ,
    local_dim=2,
    convergence_parameters=QCConvergenceParameters(),
    operators=QCOperators(),
    io_info=QCIO(),
    observables=TNObservables(),
    transpilation_parameters=qk_transpilation_params(),
    checkpoints=QCCheckpoints(),
    where_barriers=-1,
    backend=QCBackend(),
    cmd_args=None,
):
    """
    Transpile the circuit to adapt it to the linear structure of the MPS and run the circuit,
    obtaining in output the measurements.

    Parameters
    ----------
    circ: QuantumCircuit or strawberryfields.Program
        qiskit quantum circuit object to simulate
    local_dim: int, optional
        Local dimension of the single degree of freedom. Default is 2, for qubits
    convergence_parameters: :py:class:`QCConvergenceParameters`, optional
        Maximum bond dimension and cut ratio. Default to max_bond_dim=10, cut_ratio=1e-9.
    operators: :py:class:`QCOperators`, optional
        Operator class with the observables operators ALREADY THERE. If None, then it is
        initialized empty. Default to None.
    io_info: :py:class:`QCIO`, optional
        Informations about input/output files to interface with the FORTRAN simulator
    observables: :py:class:`TNObservables`, optional
        The observables to be measured at the end of the simulation. Default to TNObservables(),
        which contains no observables to measure.
    transpilation_parameters: :py:class:`qk_transpilation_params`, optional
        Parameters used in the qiskit transpilation phase. Default to qk_transpilation_params().
    checkpoints: :py:class:`QCCheckpoints`, optional
        Class to handle checkpoints in the simulation
    where_barriers: int, optional
        This parameter is important only if you want to use MPI parallelization.
        If where_barriers > 0 then the circuit gets ordered in layers and a barrier is applyed
        each where_barriers layers. We recall that a barrier is equivalent to a canonization in
        the MPS simulation. Default to -1.
    backend: :py:class:`QCBackend`, optional
        Backend containing all the information for where to run the simulation
    cmd_args: list of strings, optional
        Allows controlling the arguments parsed by argparse within qmatchatea.
        Default to `None` which retrieves arguments from sys.argv

    Returns
    -------
    result: qmatchatea.simulation_results
        Results of the simulation, containing the following data:
        - Measures
        - Statevector
        - Computational time
        - Singular values cut
        - Entanglement
        - Measure probabilities
        - MPS state
        - Observables measurements
    """
    # Checks on input parameters
    if not isinstance(convergence_parameters, QCConvergenceParameters):
        raise TypeError(
            "convergence_parameters must be of type QCConvergenceParameters"
        )
    elif not isinstance(operators, QCOperators):
        raise TypeError("operators must be of type QCOperators")
    elif not isinstance(io_info, QCIO):
        raise TypeError("io_info must be of type QCIO")
    elif not isinstance(observables, TNObservables):
        raise TypeError("observables must be of type TNObservables")
    elif not isinstance(transpilation_parameters, qk_transpilation_params):
        raise TypeError(
            "transpilation_parameters must be of type qk_transpilation_params"
        )
    elif not isinstance(checkpoints, QCCheckpoints):
        raise TypeError("checkpoints must be of type QCCheckpoints")
    elif not isinstance(backend, QCBackend):
        raise TypeError("backend must be of type QCBackend")

    # Checks on command line arguments that MIGHT modify the script
    io_info, backend, where_barriers = _parse_and_modify(
        cmd_args, io_info, backend, where_barriers
    )

    # If the approach is pure python, do not go through all the steps for
    # the usual simulation setting
    if backend.backend == "PY":
        return run_py_simulation(
            circ,
            local_dim,
            convergence_parameters,
            operators,
            observables,
            transpilation_parameters=transpilation_parameters,
            backend=backend,
            initial_state=io_info.initial_state,
        )

    # Ensure observables output folders is present and set IO
    io_info.setup()
    if backend.num_procs > 1:
        new_exe = ["mpiexec", "-n", str(backend.num_procs), *io_info.exe_file]
        io_info.set_exe_file(new_exe)
    if not os.path.isdir(os.path.join(io_info.outPATH, observables.folder_observables)):
        os.makedirs(os.path.join(io_info.outPATH, observables.folder_observables))

    # Set the PATH to the saved files into the output folder
    for ii in range(len(observables.obs_list["TNState2File"])):
        if "/" not in observables.obs_list["TNState2File"].name[ii]:
            observables.obs_list["TNState2File"].name[ii] = os.path.join(
                io_info.outPATH, observables.obs_list["TNState2File"].name[ii]
            )

    # Checks for parallel implementation
    if backend.mpi_approach not in ("SR", "PF"):
        if io_info.exe_cmd[0] == EXE_PATH:
            raise RuntimeError("Selected parallel approach using serial executable")
        if where_barriers <= 0:
            raise ValueError(
                "To obtain a correct result in a parallel approach the \
                where_barriers parameters must be greater then 0"
            )
        if backend.mpi_approach == "MW":
            if where_barriers != 1:
                raise ValueError(
                    "To obtain a correct result with the MW approach \
                    where_barrier must be 1."
                )

    # Preprocess the circuit to adapt it to the MPS constraints (linearity)
    preprocessed_circ = preprocess(circ, qk_params=transpilation_parameters)

    # If required order the circuit in a layered way and apply barriers
    if where_barriers > 0:
        preprocessed_circ = to_layered_circ(
            preprocessed_circ, where_barriers=where_barriers
        )

    # Write the circuit on file
    n_sites, operator_mapping, circ_str = pre_parser(
        preprocessed_circ,
        operators,
        io_info.inPATH,
        fock_cutoff=local_dim,
        sparse=io_info.sparse,
    )

    # Write the observables on file
    observables.write(io_info.inPATH, {}, operator_mapping)

    # Prepare input dictionary
    input_dict = OrderedDict()
    input_dict["num_sites"] = n_sites
    input_dict["local_dim"] = local_dim
    input_dict["observables_filename"] = os.path.join(observables.filename_observables)

    dictionaries = [
        input_dict,
        convergence_parameters.to_dict(),
        io_info.to_dict(),
        checkpoints.to_dict(),
        backend.to_dict(),
        OrderedDict({"state_string_len": 0}),
    ]
    temp_input_dict = merge_ordered_dicts(dictionaries)
    # Add "settings%" on front for ease fortran reading
    input_dict = OrderedDict()
    for key, val in temp_input_dict.items():
        input_dict[f"settings%{key}"] = val

    # Save input parameters on namelist file
    write_nml("INPUT", input_dict, os.path.join(io_info.inPATH, io_info.input_namelist))

    # Setting checkpoints if required
    if checkpoints.frequency > 0:
        checkpoints.set_up(input_dict, operators, observables, circ_str)

    # Run the fortran program with the namelist name as argument
    if backend.mpi_approach == "PF":
        res = mock_fortran(io_info.exe_cmd, observables, operator_mapping)
    else:
        res = subprocess.run(io_info.exe_file, capture_output=True)
        if res.stderr:
            print("--------------------------------------------")
            print("Program terminated with the following error:")
            print(res.stderr.decode())
            print("--------------------------------------------")
            sys.exit(1)
    result = simulation_results(input_dict, observables)

    # Retrieve results
    result._get_results()

    return result


def restart_from_checkpoint(io_info, checkpoint_number=-1):
    """
    Restart a simulation starting from a checkpoint. Remember not to move checkpoints around,
    since it is expected the following directory tree

    .. code-block::bash
        checkpoints/
        |-general_setting_files
        |-checkpoint_1/
        |-...
        ...
        |-checkpoint_n/

    Parameters
    ----------
    io_info : :py:class:`QCIO`
        IO information for the checkpoints. The INPATH is the folder where
        the checkpoints are saved, the OUTPATH is the folder where the results
        of the simulation, starting from that checkpoint, will be saved.
    checkpoint_number: int
        Number of the directory you want to restart from. If `-1` restart from the last one.
        Default to -1.

    Returns
    -------
    result: qmatchatea.simulation_results
        Results of the simulation
    """
    if not isinstance(io_info, QCIO):
        raise TypeError("io_info must be of type QCIO")

    PATH = io_info.inPATH
    if not os.path.isdir(PATH):
        raise IOError("Selected checkpoint directory does not exists")

    # Search all the directories in PATH that have 'checkpoint' in the name
    checkpoints = np.array(
        [
            o
            for o in os.listdir(PATH)
            if (os.path.isdir(os.path.join(PATH, o)) and "checkpoint" in o)
        ]
    )

    if len(checkpoints) == 0:
        raise RuntimeError("No checkpoints founded in the given PATH")
    # Order them by the number after the '_'
    checkpoints_numbers = [int(ck[ck.rfind("_") + 1 :]) for ck in checkpoints]
    order = np.argsort(checkpoints_numbers)
    checkpoints = checkpoints[order]
    # Pick the one selected by the user
    checkpoint_dir = str(checkpoints[checkpoint_number])
    # Select file
    checkpoint_file = os.listdir(os.path.join(PATH, checkpoint_dir))[0]
    checkpoint_file = os.path.join(checkpoint_dir, checkpoint_file)
    # The trailing number in the checkpoint file is the starting line for the circuit file
    starting_line = int(checkpoint_file[checkpoint_file.rfind("_") + 1 :])

    io_info.setup()
    io_info.set_initial_state(checkpoint_file)

    # Retrieve input dictionary
    namelist_name, input_dict = read_nml(os.path.join(PATH, io_info.input_namelist))
    # Retrieve observables class
    with open(os.path.join(PATH, "observables.pk"), "rb") as fh:
        observables = pickle.load(fh)
    obs_out_path = os.path.join(io_info.outPATH, observables.folder_observables)
    if not os.path.isdir(obs_out_path):
        os.mkdir(obs_out_path)
    # Modify the input dict to add the new informations
    input_dict["settings%initial_state"] = checkpoint_file
    input_dict["settings%initial_line"] = starting_line
    input_dict["settings%outPATH"] = io_info.outPATH

    # Save input parameters on namelist file
    write_nml(namelist_name, input_dict, os.path.join(PATH, io_info.input_namelist))

    # Run the fortran program with the namelist name as argument
    res = subprocess.run(io_info.exe_file, capture_output=True)
    if res.stderr:
        print("--------------------------------------------")
        print("Program terminated with the following error:")
        print(res.stderr.decode())
        print("--------------------------------------------")
        sys.exit(1)
    result = simulation_results(input_dict, observables)

    # Retrieve results
    result._get_results()

    return result


def _parse_and_modify(cmd_args, io_info, backend, where_barriers):
    """
    Parse command line input and modify at runtime the change of approach.
    If the --approach command line argument is present, it overwrites the script approach.
    If --approach is absent, but --mpi is present, the script approach is overwritten to 'CT'

    Parameters
    ----------
    cmd_args: list of strings or `None`
        Allows controlling the arguments parsed by argparse within qmatchatea.
        Passing `None` retrieves arguments from sys.argv
    io_info : :class: QCIO
        Input information. The executable path might be modified
    approach : QCBackend
        Approach of the simulation. If command lines are present might be modified
    where_barriers : int
        Where to put barriers in the parallel approach. If it is -1 and the command
        line arguments are present, it is modified to 1.

    Returns
    -------
    :class: QCIO
        New io information
    str
        New approach
    where_barriers
        New number of barriers
    """
    parser = argparse.ArgumentParser(
        prog="qmatchatea parser",
        description="Change the behavior of your script at runtime",
    )
    parser.add_argument(
        "--mpi",
        help="Number of cores to be used. Set the approach to"
        + "'CT' if not otherwise stated",
    )
    parser.add_argument(
        "--approach", help="Modify at runtime the approach used in the simulation"
    )

    # Dummy argument for running in jupyter
    parser.add_argument("-f")
    args, _ = parser.parse_known_args(cmd_args)

    # If approach is present, overwrite the script approach
    if args.approach:
        approach = args.approach
        backend._mpi_approach = approach
        # Set the correct file if approach is serial
        if approach == "SR":
            new_exe = [EXE_PATH]
            io_info.set_exe_file(new_exe)
    # If approach is absent and mpi is present, approach becomes 'CT'
    elif args.mpi:
        backend._mpi_approach = "CT"

    # If mpi is present we modify the io info to add the parallel run
    if args.mpi:
        backend._num_procs = int(args.mpi)
        new_exe = "mpirun -np " + str(args.mpi) + " " + EXE_PATH
        io_info.set_exe_file(new_exe.split())

    # Modify where_barriers if any argument is present and where_barriers is -1
    if (args.approach or args.mpi) and where_barriers == -1:
        where_barriers = 1

    return io_info, backend, where_barriers



import os
import os.path
import unittest
from shutil import rmtree
import subprocess

try:
    import mpi4py

    mpi_is_present = True
except ModuleNotFoundError:
    mpi_is_present = False


class TestSimulationParallelization(unittest.TestCase):
    def setUp(self):
        if not os.path.isdir("TMP_TEST"):
            os.makedirs("TMP_TEST")
        self.exe_is_present = os.path.isfile("main_qmatchatea.exe")

    def tearDown(self):
        if os.path.isdir("TMP_TEST"):
            rmtree("TMP_TEST")

    # def test_exe_presence(self):
    #    """
    #    Test if the serial executable is present
    #    """
    #    self.assertTrue(self.exe_is_present, 'Executable file is present')
    #    return

    @unittest.skipIf(not mpi_is_present, "MPI is nor present")
    def test_mpi4py(self):
        """
        Test if the parallelization of multiple simulations is working using GHZ states.
        Since it is not possible to directly run this test we perform a workaround:
        through subprocess the unittest calls the actual test
        """
        if not self.exe_is_present:
            return

        result = subprocess.run(
            ["mpiexec", "-n", "4", "python3", "python/tests/_testmpi.py"],
            capture_output=True,
        )

        out = str(result.stdout).split("\n")
        condition = bool(out[-1])

        self.assertTrue(
            condition,
            msg="GHZ circuits simulated parallely on multiple processes correts",
        )

        return



# This code is part of qmatchatea.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""
MPS python simulator.
Mimicks exactly the behavior of the FORTRAN simulator if it is
called with the :py:func:`run_simulation`, reading from files.

It can also enable full python simulations without the IO interface
by using the :py:class:`QCEmulator` class.

Functions and classes
~~~~~~~~~~~~~~~~~~~~~

"""

import os
import time
from warnings import warn
from copy import deepcopy

import numpy as np
from qiskit import QuantumCircuit
from qtealeaves import read_tensor
from qtealeaves.convergence_parameters import TNConvergenceParameters
from qtealeaves.emulator import MPS, TTN, _AbstractTN
from qtealeaves.observables import TNObservables

# Try to import cupy
try:
    import cupy as cp
    from cupy_backends.cuda.api.runtime import CUDARuntimeError

    try:
        _ = cp.cuda.Device()
        GPU_AVAILABLE = True
    except CUDARuntimeError:
        GPU_AVAILABLE = False
except ImportError:
    cp = None
    GPU_AVAILABLE = False

from .circuit import Qcircuit
from .circuit.observables import QCObservableStep
from .preprocessing import preprocess
from .tn_utils import read_nml, QCOperators, write_tensor
from .qk_utils import qk_transpilation_params
from .utils import QCIO, QCConvergenceParameters, simulation_results
from .utils import QCBackend

__all__ = ["QcMps", "QCEmulator", "mock_fortran", "run_py_simulation"]


class QCEmulator:
    """
    Emulator class to run quantum circuits, powered by either
    TTNs or MPS.


    Parameters
    ----------

    num_sites: int
        Number of sites
    num_clbits: int
        Number of classical bits
    convergence_parameters: :py:class:`QCConvergenceParameters`
        Class for handling convergence parameters. In particular, in the MPS simulator we are
        interested in:
        - the *maximum bond dimension* :math:`\\chi`;
        - the *cut ratio* :math:`\\epsilon` after which the singular values are neglected, i.e.
          if :math:`\\lambda_1` is the bigger singular values then after an SVD we neglect all the
          singular values such that :math:`\\frac{\\lambda_i}{\\lambda_1}\\leq\\epsilon`
    local_dim: int, optional
        Local dimension of the degrees of freedom. Default to 2.
    """

    def __init__(
        self,
        num_sites,
        convergence_parameters=QCConvergenceParameters(),
        local_dim=2,
        dtype=np.complex128,
        device="cpu",
        ansatz="MPS",
    ):
        if not isinstance(convergence_parameters, TNConvergenceParameters):
            raise TypeError(
                "convergence_parameters must be of the QCConvergenceParameters class"
            )

        self._trunc_tracking_mode = convergence_parameters.trunc_tracking_mode
        self.ansatz = ansatz

        # Classical registers to hold qiskit informations
        self.cl_regs = {}

        # Observables measured
        self.is_measured = [
            True,
            False,
            False,
            True,
            True,
            True,
            True,
            True,
            True,
            False,
            False,
        ]

        if ansatz.upper() == "MPS":
            self.emulator = MPS(
                num_sites=num_sites,
                convergence_parameters=convergence_parameters,
                local_dim=local_dim,
                dtype=dtype,
                device=device,
            )
        elif ansatz.upper() == "TTN":
            zero_state = np.zeros((num_sites, local_dim), dtype=dtype)
            zero_state[:, 0] = 1
            self.emulator = TTN.product_state_from_local_states(
                zero_state,
                convergence_parameters=convergence_parameters,
                dtype=dtype,
                device=device,
            )
        else:
            raise ValueError(f"Ansatz {ansatz} is not available for QCEmulator")

    def __getattr__(self, __name: str):
        """
        Check for the attribute in emulator, i.e. the QCEmulator inherits all
        the emulator calls.
        This call is for convenience and for retrocompatibility

        .. warning::
            The method `__getattr__` is called when `__getattribute__` fails,
            so it already covers the possibility of the attribute being in the
            base class
        """
        return self.emulator.__getattribute__(__name)

    @classmethod
    def from_emulator(cls, emulator, conv_params=None, dtype=None):
        """
        Initialize the QCEmulator class starting from an emulator class, i.e. either
        MPS or TTN

        Parameters
        ----------
        emulator : :class:`_AbstractTN`
            Either an MPS or TTN emulator
        conv_params : :class:`TNConvergenceParameters`, optional
            Convergence parameters. If None, the convergence parameters of the emulator
            are used
        dtype : type, optional
            Type to be used, If None, the dtype of the emulator is used

        Return
        ------
        QCEmulator
            The quantum circuit emulator class
        """
        if not isinstance(emulator, _AbstractTN):
            raise TypeError("The emulator should be a TN emulator class")
        if conv_params is None:
            conv_params = emulator.convergence_parameters
        if dtype is None:
            dtype = emulator.dtype

        simulator = cls(
            emulator.num_sites, conv_params, emulator.local_dim, dtype=dtype
        )
        simulator.emulator = emulator
        simulator.ansatz = str(emulator)

        return simulator

    @classmethod
    def from_tensor_list(cls, tensor_list, conv_params=None, dtype=None):
        """
        Initialize the QCEmulator class starting from a tensor list, i.e. either
        MPS or TTN

        Parameters
        ----------
        tensor_list : list of tensors
            Either an MPS or TTN list of tensors
        conv_params : :class:`TNConvergenceParameters`, optional
            Convergence parameters. If None, the convergence parameters of the emulator
            are used
        dtype : type, optional
            Type to be used, If None, the dtype of the emulator is used

        Return
        ------
        QCEmulator
            The quantum circuit emulator class
        """
        # A list of lists is a TTN, while a list of tensors is an MPS
        if isinstance(tensor_list[0], list):
            initial_state = TTN.from_tensor_list(tensor_list)
        else:
            initial_state = MPS.from_tensor_list(tensor_list)

        simulator = cls.from_emulator(
            initial_state, conv_params=conv_params, dtype=dtype
        )

        return simulator

    def meas_projective(
        self, nmeas=1024, qiskit_convention=True, seed=None, unitary_setup=None
    ):
        """See the parent method"""
        return self.emulator.meas_projective(
            nmeas=nmeas,
            qiskit_convention=qiskit_convention,
            seed=seed,
            unitary_setup=unitary_setup,
        )

    def to_statevector(self, qiskit_order=True, max_qubit_equivalent=20):
        """See the parent method"""
        return self.emulator.to_statevector(qiskit_order, max_qubit_equivalent)

    def apply_two_site_gate(self, operator, control, target):
        """Apply a two-site gate, regardless of the position on the chain

        Parameters
        ----------
        operator : np.ndarray
            Gate to be applied
        control : int
            control qubit index
        target : int
            target qubit index

        Returns
        -------
        singvals_cut
            singular values cut in the process
        """
        xp = self._device_checks()
        if operator.shape == (4, 4):
            operator = operator.reshape(2, 2, 2, 2)
        # Reorder for qiskit convention on the two-qubits gates
        if control < target or self.ansatz == "TTN":
            operator = xp.transpose(operator, [1, 0, 3, 2])

        singvals_cut = self.apply_two_site_operator(operator, [control, target])

        # Avoid errors due to no singv cut
        singvals_cut = np.append(singvals_cut, 0)
        if self._trunc_tracking_mode == "M":
            singvals_cut = max(0, xp.max(singvals_cut))
        elif self._trunc_tracking_mode == "C":
            singvals_cut = xp.sum(singvals_cut**2)

        return [singvals_cut]

    def meas_observables(self, observables, op_list, op_mapping=None, approach="PF"):
        """Measure all the observables

        Parameters
        ----------
        observables : :py:class:`TNObservables`
            All the observables to be measured
        op_list : list of tensors
            List of operators that form the circuit stored in THE CORRECT DEVICE.
            If you are running on GPU the operators should be on the GPU.
        op_mapping : dict
            Mapping between the operators name and the operators idx
        approach : string, optional
            Approach of the simulation, if 'PY', 'PF'

        Returns
        -------
        TNObservables
            Observables with the results in results_buffer
        """
        if not isinstance(observables, TNObservables):
            raise TypeError("observables must be TNObservables")

        keys = np.array(list(observables.obs_list.keys()))[self.is_measured]
        ii = 0
        ## LOCAL OBSERVABLES ##
        local = observables.obs_list[keys[ii]]
        for name, idx in zip(local.name, local.operator):
            if approach == "PF":
                idx = op_mapping[idx] - 1
            local.results_buffer[name] = self.meas_local(op_list[idx])
        ii += 1

        ## SAVE STATE ##
        state_obs = observables.obs_list[keys[ii]]
        if len(state_obs.name) > 0:
            if approach == "PY":
                observables.results_buffer["tn_state"] = self.emulator.to_tensor_list()
            else:
                name = state_obs.name[0]
                state_obs.results_buffer[name] = name
                self.emulator.write(name)
        ii += 1

        ## TENSOR PRODUCT OBSERVABLES ##
        tensor_prod = observables.obs_list[keys[ii]]
        for name, ops, sites in zip(
            tensor_prod.name, tensor_prod.operators, tensor_prod.sites
        ):
            sites = [site[0] for site in sites]
            if approach == "PF":
                operators = [op_list[op_mapping[ii] - 1] for ii in ops]
            else:
                operators = [op_list[ii] for ii in ops]
            tensor_prod.results_buffer[name] = np.complex128(
                self.emulator.meas_tensor_product(operators, sites)
            )
        ii += 1

        ## WEIGHTED SUM OBSERVABLES ##
        wsum = observables.obs_list[keys[ii]]
        for name, coef, tp_ops in zip(wsum.name, wsum.coeffs, wsum.tp_operators):
            op_string = []
            idxs_string = []
            if isinstance(tp_ops, list):
                tp_op = tp_ops[0]
            else:
                tp_op = tp_ops
            for ops, sites in zip(tp_op.operators, tp_op.sites):
                sites = [site[0] for site in sites]
                if approach == "PF":
                    operators = [op_list[op_mapping[ii] - 1] for ii in ops]
                else:
                    operators = [op_list[ii] for ii in ops]
                idxs_string.append(sites)
                op_string.append(operators)

            wsum.results_buffer[name] = np.complex128(
                self.emulator.meas_weighted_sum(op_string, idxs_string, coef)
            )
        ii += 1

        ## FINAL PROJECTIVE MEASUREMENT OBSERVABLES ##
        proj_obs = observables.obs_list[keys[ii]]
        proj_obs.results_buffer[proj_obs.name[0]] = self.meas_projective(
            nmeas=proj_obs.num_shots
        )
        ii += 1

        ## PROBABILITY MEASURE ##
        prob_obs = observables.obs_list[keys[ii]]
        for name, prob_type, prob_param in zip(
            prob_obs.name, prob_obs.prob_type, prob_obs.prob_param
        ):
            if isinstance(name, list):
                name = name[0]
            if prob_type == "U":
                prob_obs.results_buffer[
                    name
                ] = self.emulator.meas_unbiased_probabilities(prob_param, True)
            elif prob_type == "E":
                prob_obs.results_buffer[name] = self.emulator.meas_even_probabilities(
                    prob_param, True
                )
            elif prob_type == "G":
                prob_obs.results_buffer[name] = self.emulator.meas_greedy_probabilities(
                    prob_param, True
                )
        ii += 1

        ## ENTANGLEMENT MEASURE ##
        ent_obs = observables.obs_list[keys[ii]]
        if len(ent_obs.name) == 1:
            bond_entropy = self.emulator.meas_bond_entropy()
            ent_obs.results_buffer[ent_obs.name[0]] = bond_entropy
        ii += 1

        return observables

    def run_circuit_from_instruction(self, op_list, instr_list):
        """Run a circuit

        Parameters
        ----------
        op_list : list of tensors
            List of operators that form the circuit
        instr_list : list of instructions
            Instruction for the circuit, i.e. [op_name, op_idx, [sites] ]

        Return
        ------
        singvals_cut : list of float
            Singular values cutted, selected through the _trunc_tracking_mode
        """
        singvals_cut = []
        for instr in instr_list:
            sites = instr[2]
            num_sites = len(sites)
            idx = instr[1]
            if instr[0] == "barrier":
                continue

            if num_sites == 1:
                self.emulator.apply_one_site_operator(op_list[idx], *sites)

            elif num_sites == 2:
                singv_cut = self.apply_two_site_gate(op_list[idx], sites[0], sites[1])

                # Avoid errors due to no singv cut
                singv_cut = np.append(singv_cut, 0)
                if self._trunc_tracking_mode == "M":
                    singvals_cut.append(np.max(singv_cut, initial=0.0))
                elif self._trunc_tracking_mode == "C":
                    singvals_cut.append(np.sum(singv_cut**2))

            else:
                raise ValueError("Only one and two-site operations are implemented")
        return singvals_cut

    def run_from_qk(self, circuit):
        """
        Run a qiskit quantum circuit on the simulator

        Parameters
        ----------
        circuit : :py:class:`QuantumCircuit`
            qiskit quantum circuit

        Returns
        -------
        list of floats
            singular values cutted in the simulation
        """
        xp = self._device_checks()
        data = circuit.data  # data structure of the quantum circuit
        singvals_cut = []
        for creg in circuit.cregs:
            self.cl_regs[creg.name] = np.zeros(creg.size)

        # Run over instances
        for instance in data:
            gate_name = instance[0].name
            num_qubits = len(instance[1])
            qubits = [circuit.find_bit(qub).index for qub in instance[1]]

            if gate_name == "barrier":
                continue
            elif gate_name == "measure":
                meas_state, _ = self.apply_projective_operator(*qubits)
                self.cl_regs[instance[2][0].register.name][0] = meas_state
                continue
            elif gate_name == "reset":
                self.reset(qubits)
                continue
            else:
                gate_mat = instance[0].to_matrix()

            if instance[0].condition is None:
                apply_gate = True
            else:
                bit_idx = [clbit.index for clbit in instance[0].condition[0]]
                bit_value = self.cl_regs[instance[0].condition[0].name][bit_idx[0]]
                apply_gate = bit_value == instance[0].condition[1]

            if apply_gate:
                # Move the operator to the GPU
                if xp == cp:
                    gate_mat = cp.asarray(gate_mat)
                if num_qubits == 1:
                    self.emulator.apply_one_site_operator(gate_mat, *qubits)
                elif num_qubits == 2:
                    singv_cut = self.apply_two_site_gate(gate_mat, *qubits)
                    singvals_cut += singv_cut
                else:
                    raise ValueError("Only one and two-site operations are implemented")
        return singvals_cut

    def run_from_qcirc(self, qcirc, starting_idx=0):
        """
        Run a simulation starting from a Qcircuit on a portion of the MPS state

        Parameters
        ----------
        qcirc : :class:`Qcircuit`
            Quantum circuit
        starting_idx : int, optional
            MPS index that correspond to the index 0 of the Qcircuit. Default to 0.
        """
        if not isinstance(qcirc, Qcircuit):
            raise TypeError(f"qcirc must be of type Qcircuit, not {type(qcirc)}")
        xp = self._device_checks()
        results = {}
        svd_cuts = []
        start_time = time.time()
        for layer in qcirc:
            for instruction in layer:
                sites = [ss + starting_idx for ss in instruction[1]]
                operation = instruction[0]

                # Check for classical conditioning
                appy_operation = operation.c_if.is_satisfied(qcirc)
                if appy_operation:
                    # First, check for particular keywords
                    if isinstance(operation, QCObservableStep):
                        obs = self.meas_observables(
                            operation.observables,
                            operation.operators.ops,
                            approach="PY",
                        )
                        obs.results_buffer["time"] = time.time() - start_time
                        obs.results_buffer["norm"] = self.norm()
                        operation.observables = obs
                        operation.postprocess_obs_indexing()  # Postprocess for qregisters
                        for elem in obs.obs_list:
                            obs.results_buffer.update(obs.obs_list[elem].results_buffer)
                        results[operation.name] = deepcopy(
                            operation.observables.results_buffer
                        )
                        del obs

                    # Check for particular keywords
                    elif operation.name == "renormalize":
                        self.normalize()
                    elif operation.name == "measure":
                        res = self.emulator.apply_projective_operator(
                            *sites, operation.selected_output
                        )
                        # Update measured value
                        qcirc.modify_cregister(
                            res, operation.cregister, operation.cl_idx
                        )
                    elif operation.name == "add_site":
                        self.emulator.add_site(operation.position)
                    elif operation.name == "remove_site":
                        self.apply_projective_operator(operation.position, remove=True)

                    # Apply gates
                    elif len(sites) == 1:
                        if xp == cp:
                            gate_mat = cp.asarray(gate_mat)
                        self.site_canonize(*sites, keep_singvals=True)
                        self.apply_one_site_operator(operation.operator, *sites)
                    else:
                        if xp == cp:
                            gate_mat = cp.asarray(gate_mat)
                        svd_cut = self.apply_two_site_gate(operation.operator, *sites)
                        svd_cuts += svd_cut

        # Truncation
        results["singular_values_cut"] = svd_cuts
        return results


class QcMps(QCEmulator):
    """
    Class for retrocompatibility with QCEmulator
    """

    def __init__(self, *args, **kwargs):
        warn("Using deprecated QcMps, please switch to QCEmulator")
        super().__init__(*args, **kwargs)


def mock_fortran(command_line, observables, op_mapping):
    """
    Mock the execution of the Fortran program, by doing the exact same things Fortran would do

    Parameters
    ----------
    command_line : list of str
        List of commands that should be run on terminal for the compiled fortran simulator
    """
    input_dict_path = command_line[-1]
    if not os.path.isfile(input_dict_path):
        raise ValueError(
            'The given PATH "%s" for the input file is not a file.'
            % (str(input_dict_path))
        )

    _, input_dict = read_nml(command_line[-1])

    # Checks and retrieval on input parameters
    if input_dict["checkpoint_frequency"] > 0:
        raise NotImplementedError(
            "Checkpoints are not implemented in the python simulator"
        )
    io_info = QCIO(
        inPATH=input_dict["inPATH"],
        outPATH=input_dict["outPATH"],
        sparse=input_dict["sparse"],
    )
    conv_params = QCConvergenceParameters(
        max_bond_dimension=input_dict["max_bond_dimension"],
        cut_ratio=input_dict["cut_ratio"],
        trunc_tracking_mode=input_dict["trunc_tracking_mode"],
    )

    # Retrieve input information on operators and circuit
    op_list, instruction_list, num_sites, num_cl_bits = _mock_fortran_input(io_info)

    # Initialize the state
    emulator = QCEmulator(
        num_sites,
        num_cl_bits,
        conv_params,
        input_dict["local_dim"],
        ansatz=input_dict["ansatz"],
    )

    # Retrieve initial state if present
    if input_dict["initial_state"] != "Vacuum":
        initial_state_path = os.path.join(io_info.inPATH, input_dict["initial_state"])
        emulator.read(initial_state_path)

    # Run the circuit
    start = time.time()
    singvals_cut = emulator.run_circuit_from_instruction(op_list, instruction_list)
    simulation_time = time.time() - start

    # Perform measurements and final operations
    observables = _mock_fortran_measurements(emulator, observables, op_list, op_mapping)
    observables.results_buffer["time"] = simulation_time

    # Write everithing in the output directory
    _mock_fortran_output(emulator, io_info, input_dict, singvals_cut, observables)

    return True


def _mock_fortran_input(io_info):
    """
    Perform the reading necessary for the python emulator

    Parameters
    ----------
    io_info : :py:class:`QCIO`
        IO informations

    Returns
    -------
    op_list: list of ndarray
        List of operators
    instruction_list: list
        list of instruction for the circuit
    num_sites: int
        Number of sites in the MPS
    num_cl_bits : int
        Number of classical bits in the MPS
    """
    if not isinstance(io_info, QCIO):
        raise TypeError("io_info must be of type QCIO")

    # Retrieve operators
    op_path = os.path.join(io_info.inPATH, "TENSORS/operators.dat")
    op_list = []
    with open(op_path, "r") as fh:
        op_num = int(fh.readline().replace("\n", ""))
        for _ in range(op_num):
            tens = read_tensor(fh)
            op_list.append(tens)

    # Retrieve circuit
    instruction_path = os.path.join(io_info.inPATH, "circuit.dat")
    instruction_list = []
    with open(instruction_path, "r") as fh:
        num_sites = int(fh.readline().replace("\n", ""))
        num_cl_bits = int(fh.readline().replace("\n", ""))
        num_lines = int(fh.readline().replace("\n", ""))
        for _ in range(num_lines):
            line_1 = fh.readline().replace("\n", "").split(" ")
            line_2 = fh.readline().replace("\n", "").split(" ")
            line_2.remove("")
            # [ name, op_idx, [qubits] ]
            instruction = [
                line_1[0],
                int(line_1[1]) - 1,
                [int(site) - 1 for site in line_2],
            ]

            instruction_list.append(instruction)

    return op_list, instruction_list, num_sites, num_cl_bits


def _mock_fortran_measurements(emulator, observables, op_list, op_mapping):
    """
    Perform the different measurements to mock with python a fortran simulation

    Parameters
    ----------
    emulator : :py:class:`QCEmulator`
        Matrix product state class
    observables : :py:class:`TNObservables`
        Observables class
    op_list : list of ndarray
        List of operators

    Return
    ------
    measurements : OrderedDict
        Ordered dictionary with as keys:
            - proj_meas, the projective measurements dictionary
            - statevector, the statevector
            - entanglement, the entanglement
            - obs_results, the observables results

    """
    if not isinstance(emulator, QCEmulator):
        raise TypeError("mps_state must be of type emulator")
    elif not isinstance(observables, TNObservables):
        raise TypeError("observables must be of type TNObservables")

    # Observables with measured quantities
    observables = emulator.meas_observables(observables, op_list, op_mapping)
    observables.results_buffer["energy"] = 999e300
    observables.results_buffer["norm"] = emulator.norm()

    return observables


def _mock_fortran_output(emulator, io_info, input_dict, singval_cut, observables):
    """
    Mock the fortran simulator by writing the same output files it would
    write

    Parameters
    ----------
    emulator : :py:class:`QCEmulator`
        QCEmulator state class
    io_info : :py:class:`QCIO`
        Info about the output file path
    input_dict : OrderedDict
        Ordered dictionary with the input parameters
    singval_cut : array-like
        Singular values cut in the simulation
    observables : :py:class:`TNObservables`
        Observables
    """
    if not isinstance(emulator, QCEmulator):
        raise TypeError("mps_state must be of type QCEmulator")
    elif not isinstance(io_info, QCIO):
        raise TypeError("io_info must be of type QCIO")

    # Observables
    obs_path = os.path.normpath(input_dict["observables_filename"])
    obs_path = os.path.join(io_info.outPATH, obs_path)
    observables.write_results(obs_path, emulator.is_measured, {})

    with open(os.path.join(io_info.outPATH, "singular_values_cut.txt"), "w") as fh:
        _write_value(singval_cut, fh)


def run_py_simulation(
    circ,
    local_dim=2,
    convergence_parameters=QCConvergenceParameters(),
    operators=QCOperators(),
    observables=TNObservables(),
    initial_state=None,
    transpilation_parameters=qk_transpilation_params(),
    backend=QCBackend(),
):
    """
    Transpile the circuit to adapt it to the linear structure of the MPS and run the circuit,
    obtaining in output the measurements.

    Parameters
    ----------
    circ: QuantumCircuit or strawberryfields.Program
        qiskit quantum circuit object to simulate
    local_dim: int, optional
        Local dimension of the single degree of freedom. Default is 2, for qubits
    convergence_parameters: :py:class:`QCConvergenceParameters`, optional
        Maximum bond dimension and cut ratio. Default to max_bond_dim=10, cut_ratio=1e-9.
    operators: :py:class:`QCOperators`, optional
        Operator class with the observables operators ALREADY THERE. If None, then it is
        initialized empty. Default to None.
    observables: :py:class:`TNObservables`, optional
        The observables to be measured at the end of the simulation. Default to TNObservables(),
        which contains no observables to measure.
    initial_state : list on ndarray, optional
        Initial state of the simulation. If None, ``|00...0>`` is considered. Default to None.
    transpilation_parameters: :py:class:`qk_transpilation_params`, optional
        Parameters used in the qiskit transpilation phase. Default to qk_transpilation_params().
    backend: :py:class:`QCBackend`, optional
        Backend containing all the information for where to run the simulation

    Returns
    -------
    result: qmatchatea.simulation_results
        Results of the simulation, containing the following data:
        - Measures
        - Statevector
        - Computational time
        - Singular values cut
        - Entanglement
        - Measure probabilities
        - MPS state
        - MPS file size
        - Observables measurements
    """
    # Preprocess the circuit to adapt it to the MPS constraints (linearity)
    preprocessed_circ = preprocess(circ, qk_params=transpilation_parameters)

    if isinstance(circ, QuantumCircuit):
        num_qubits = circ.num_qubits
        # num_clbits = circ.num_clbits
    else:
        raise TypeError(
            "Only qiskit Quantum Circuits are implemented for pure python"
            + f" simulation, not {type(circ)}"
        )
    tn_type = np.complex128 if backend.precision == "Z" else np.complex64
    start = time.time()
    if initial_state is None or initial_state == "Vacuum":
        simulator = QCEmulator(
            num_qubits,
            convergence_parameters,
            local_dim=local_dim,
            dtype=tn_type,
            ansatz=backend.ansatz,
        )
    elif isinstance(initial_state, _AbstractTN):
        simulator = QCEmulator.from_emulator(
            initial_state, conv_params=convergence_parameters, dtype=tn_type
        )
    else:
        simulator = QCEmulator.from_tensor_list(
            initial_state, conv_params=convergence_parameters, dtype=tn_type
        )

    simulator.to_device(backend.device)
    singvals_cut = simulator.run_from_qk(preprocessed_circ)
    end = time.time()

    observables = simulator.meas_observables(observables, operators.ops, approach="PY")
    observables.results_buffer["time"] = end - start
    observables.results_buffer["energy"] = None
    observables.results_buffer["norm"] = simulator.norm()

    result_dict = observables.results_buffer
    for elem in observables.obs_list:
        result_dict.update(observables.obs_list[elem].results_buffer)

    results = simulation_results()
    results.set_results(result_dict, singvals_cut)

    return results


def _write_value(value, fh):
    """
    Print a value for the mock fortran output
    based on its type

    Parameters
    ----------
    value : scalar, dict or array-like
        value to be saved
    fh : file handler
        Where to write the values
    """

    # None case
    if value is not None:
        # Writing scalar results
        if np.isscalar(value):
            if np.iscomplexobj(value):
                fh.write(f"{np.real(value) }, {np.imag(value) } \n")
            else:
                fh.write(f"{value} \n")

        # Writing dictionary results
        elif isinstance(value, dict):
            fh.write(f"{len(value)} \n")
            for subkey, subvalue in value.items():
                fh.write(f"{subkey} | {subvalue} \n")

        # Writing vector results
        elif len(np.array(value[0]).shape) < 2:
            fh.write(f"{len(value)} \n")
            if np.iscomplexobj(value):
                for subvalue in value:
                    fh.write(f"{np.real(subvalue) }, {np.imag(subvalue) } \n")
            else:
                for subvalue in value:
                    fh.write(f"{subvalue} \n")
        else:
            fh.write(f"{len(value)} \n")
            for subvalue in value:
                write_tensor(subvalue, fh, np.iscomplexobj(subvalue))



import unittest
import subprocess
import numpy as np

try:
    import mpi4py

    has_mpi = True
except ImportError:
    has_mpi = False


class TestExamples(unittest.TestCase):
    """Test that all the examples works correctly"""

    def setUp(self):
        """Define 'global' variables"""
        self.num_sites = 4
        # Set seed
        np.random.seed(123)

    def test_convergence_analysis(self):
        """
        Test the convergence analysis example
        """
        res = subprocess.run(["python3", "examples/convergence_analysis.py"])
        self.assertIsNone(res.stderr, "Example failed")

    def test_get_started(self):
        """
        Test the get started example
        """
        res = subprocess.run(["python3", "examples/get_started.py"])
        self.assertIsNone(res.stderr, "Example failed")

    @unittest.skipIf(has_mpi, "mpi4py not installed")
    def test_mpi_example(self):
        """
        Test the mpi_example example
        """
        res = subprocess.run(["mpiexec", "python3", "examples/mpi_example.py", "-n=4"])
        self.assertIsNone(res.stderr, "Example failed")

    def test_mps_ttn_comparison(self):
        """
        Test the mps_ttn_comparison example
        """
        res = subprocess.run(["python3", "examples/mps_ttn_comparison.py"])
        self.assertIsNone(res.stderr, "Example failed")

    def test_quantum_fourier_transform(self):
        """
        Test the quantum_fourier_transform example
        """
        res = subprocess.run(["python3", "examples/quantum_fourier_transform.py"])
        self.assertIsNone(res.stderr, "Example failed")

    def test_random_quantum_circuit(self):
        """
        Test the random_quantum_circuit example
        """
        res = subprocess.run(["python3", "examples/random_quantum_circuit.py"])
        self.assertIsNone(res.stderr, "Example failed")

    def test_teleportation(self):
        """
        Test the teleportation example
        """
        res = subprocess.run(["python3", "examples/teleportation.py"])
        self.assertIsNone(res.stderr, "Example failed")

    def test_variational_quantum_eigensolver(self):
        """
        Test the variational_quantum_eigensolver example
        """
        res = subprocess.run(["python3", "examples/variational_quantum_eigensolver.py"])
        self.assertIsNone(res.stderr, "Example failed")



# This code is part of qmatchatea.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""
Here we mainly retrieve the Fock representation of linear optics operators.
We furthermore added a two operations for the state preparation and a barrier insertion,
which are used internally by the Fortran.

Moreover, we provide a function to initialize a Gaussian Boson Sampling circuit,
which is a really interesting research topic, and another to compute the
occupation profile of a photonic output.

Functions and classes
---------------------

"""
from collections import OrderedDict

try:
    from strawberryfields import ops
    import strawberryfields.backends.fockbackend as fock

    SF_IS_IMPORTED = True
except ModuleNotFoundError:
    SF_IS_IMPORTED = False
import numpy as np


__all__ = ["Barrier", "optical_gates", "GBS_circ", "occupation_profile"]

if SF_IS_IMPORTED:

    class Barrier(ops.Operation):
        """
        New sf operation to apply MPS canonization during MPI protocol.
        When a barrier is set then the canonization is restored. It has no effect
        in a serial simulation. If used in a python sf simulation the application
        will fail
        """

        def __init__(self):
            super().__init__([])

else:

    class Barrier:
        """
        New sf operation to apply MPS canonization during MPI protocol.
        When a barrier is set then the canonization is restored. It has no effect
        in a serial simulation. If used in a python sf simulation the application
        will fail
        """

        def __init__(self):
            raise ImportError(
                "Strawberryfields is not imported, barriers are not available"
            )


def fock_preparation(state, trunc):
    """
    Returns the tensor to bring a vacuum state in
    a predetermined Fock state

    Parameters
    ----------
    state: int
        The Fock state level
    trunc: int
        Fock cutoff
    """

    tens = np.zeros((trunc, trunc))
    tens[state, 0] = 1

    return tens


def vacuum_preparation(trunc):
    """
    Returns the tensor to bring a vacuum state in
    a vacuum state

    Parameters
    ----------
    trunc: int
        Fock cutoff
    """

    tens = np.eye(trunc)
    return tens


def sf_barrier(trunc):
    """
    Returns a vector with all 0 and 1 in the first place,
    which is not used in the simulation. Barriers are important
    for MPI processes.

    Parameters
    ----------
    trunc: int
        Fock cutoff
    """
    tens = np.zeros((trunc, trunc))
    tens[0, 0] = 1
    return tens


if SF_IS_IMPORTED:
    optical_gates = OrderedDict()
    optical_gates["a"] = fock.ops.a
    optical_gates["Dgate"] = fock.ops.displacement
    optical_gates["Sgate"] = fock.ops.squeezing
    optical_gates["Kgate"] = fock.ops.kerr
    optical_gates["Rgate"] = fock.ops.phase
    optical_gates["BSgate"] = fock.ops.beamsplitter
    optical_gates["Vgate"] = fock.ops.cubicPhase
    # optical_gates['MZgate'] = fockops.mzgate
    optical_gates["S2gate"] = fock.ops.two_mode_squeeze
    optical_gates["CKgate"] = fock.ops.cross_kerr
    optical_gates["Fock"] = fock_preparation
    optical_gates["Vacuum"] = vacuum_preparation
    optical_gates["Barrier"] = sf_barrier
else:
    optical_gates = OrderedDict()


def GBS_circ(prog, unitary, squeeze_params=None):
    """Build a GBS circuit on a program, given the interferometer unitary

    Parameters
    ----------
    prog : strawberryfields.Program
        Strawberry fields program
    unitary: ndarray, dtype=complex
        Unitary matrix to be decomposed in the interferometer
    squeeze_params: ndarray (num_modes, 2), optional
        Parameters for the squeezing operators. Default (1, 0)

    Returns
    -------
    prog: strawberryfields.Program
        Strawberry fields program with the GBS applied
    """
    if not SF_IS_IMPORTED:
        raise ImportError("Strawberryfields library is not imported or installed.")
    num_modes = prog.num_subsystems
    interf = ops.Interferometer(unitary)
    if squeeze_params is None:
        squeeze_params = np.zeros((num_modes, 2))

    assert (
        squeeze_params.shape[0] != num_modes
    ), f"Number of squeeze parameters must be equal to number of modes m={num_modes}"
    assert (
        squeeze_params.shape[1] != 2
    ), f"Each mode should have only 2 parameters, not {squeeze_params.shape[1]}"

    commands = interf.decompose(prog.register)
    with prog.context as qub:
        # Sqeezing
        for jj in range(num_modes):
            ops.Sgate(squeeze_params[jj, 0], squeeze_params[jj, 1]) | qub[jj]
    # Interferometer
    for comm in commands:
        prog.append(comm.op, comm.reg)

    return prog


def occupation_profile(measures, density=False):
    """Compute the occupation profile of the measured state

    Parameters
    ----------
    measures: dict
        Measure of the quantum optics experiment
    density: bool
        If True normalize the number of occurrencies to 1

    Return
    ------
    occupation: ndarray, dtype=int, shape(N, 2)
        Occupation profile. The first column is the number of photons,
        the second column is the number of occurrences
    """
    keys = np.array(list(measures.keys()))

    # Polish keys to have the number of photons in the state
    number_of_photons = []
    for kk in keys:
        polished_kk = kk.replace(" ", "").split(",")
        n_phot = np.array(polished_kk, dtype=int).sum()
        number_of_photons.append(n_phot)

    occupation = {}
    # Combine all the measurements
    for ii, kk in enumerate(keys):
        n_phot = number_of_photons[ii]
        n_occ = measures[kk]
        if n_phot in occupation:
            occupation[n_phot] += n_occ
        else:
            occupation[n_phot] = n_occ

    n_phot = np.array(list(occupation.keys()))
    n_occ = np.array(list(occupation.values()), dtype=float)
    order = np.argsort(n_phot)
    n_phot = n_phot[order].reshape(-1, 1)
    n_occ = n_occ[order].reshape(-1, 1)
    if density:
        n_occ = n_occ / n_occ.sum()
    occupation = np.hstack((n_phot, n_occ))

    return occupation



# This code is part of qmatchatea.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

"""
Function to help in the MPI protocol. Here you insert barriers, which means
canonization of the state each where_barriers layers. In particular, it is
important because at each layer in the circuit description is connected a
canonization in the MPS simulation. If the canonization are not often enough
the simulation might not converge.

Functions and classes
---------------------

"""

try:
    from strawberryfields import Program

    SF_IS_IMPORTED = True
except ModuleNotFoundError:
    SF_IS_IMPORTED = False
try:
    from qiskit.visualization.utils import _get_layered_instructions
except ImportError:
    # This try except makes the code compatible also with the latest version of
    # qiskit, where the _get_layered_instructions changed path
    from qiskit.visualization.circuit._utils import _get_layered_instructions
from qiskit import QuantumCircuit
import numpy as np
from .sf_utils import Barrier

__all__ = ["to_layered_circ"]


def _to_layered_circ_qk(qc, where_barriers=1):
    """
    Transform a quantum circuit in another, where the instruction are ordered by layers.
    Apply a barrier every where_barriers layers.

    Parameters
    ----------
    qc: qiskit.QuantumCircuit
        Quantum circuit
    where_barriers: int
        Apply a barrier every where_barriers layers.

    Returns
    -------
    layered_qc: qiskit.QuantumCircuit
            Ordered quantum circuit with barriers
    """
    layered_qc = QuantumCircuit(qc.num_qubits)

    layered_instructions = _get_layered_instructions(qc)[2]
    for ii, layer in enumerate(layered_instructions):
        for instruction in layer:
            layered_qc.append(instruction.op, instruction.qargs, instruction.cargs)
        if ii % where_barriers == 0:
            layered_qc.barrier()

    return layered_qc


def _obtain_layers_sf(program):
    """
    Obtain the different layers of a sf program to optimize the MPI process

    Parameters
    ----------
    program: sf.Program
        Strawberry fields program

    Returns
    -------
    layers: list
        A list of lists. Each sublist is a layer of the program, containing the commands
    """
    if not SF_IS_IMPORTED:
        raise ImportError("Strawberryfields library is not imported or installed.")
    layers = []
    # Current_idxs contains the "position" of the last gate on each wire (qumode)
    current_idxs = np.zeros(program.num_subsystems, dtype=int)
    for com in program.circuit:
        if max(current_idxs) >= len(layers):
            layers.append([])
        idxs = [rg.ind for rg in com.reg]
        idx = min(idxs)
        layers[current_idxs[idx]].append(com)
        # You apply a gate, so the first available position shifts right (+1)
        current_idxs[idxs] += 1

    # Drop last empty layer
    layers = layers[:-1]

    return layers


def _to_layered_circ_sf(program, where_barriers=1):
    """
    Transform a strawberry fields program in another,
    where the instruction are ordered by layers.
    Apply a barrier every where_barriers layers.

    Parameters
    ----------
    program: sf.Program
        strawberry fields program
    where_barriers: int
        Apply a barrier every where_barriers layers.

    Returns
    -------
    layered_program: sf.Program
        Ordered sf program with barriers
    """
    if not SF_IS_IMPORTED:
        raise ImportError("Strawberryfields library is not imported or installed.")
    layered_program = Program(program.num_subsystems)
    layers = _obtain_layers_sf(program)

    for ii, layer in enumerate(layers):
        for comm in layer:
            idxs = [rg.ind for rg in comm.reg]
            regs = [layered_program.register[idx] for idx in idxs]

            layered_program.append(comm.op, regs)
        if ii % where_barriers == 0:
            layered_program.append(Barrier(), layered_program.register)

    return layered_program


def to_layered_circ(circ, where_barriers=1):
    """
    Interface for the transformation of qiskit and strawberry fields circuit
    into layered ones, the correct order for the MPI protocols.
    You also insert barriers, i.e. canonization procedures

    Parameters
    ----------
    circ: strawberryfields.program.Program or qiskit.circuit.quantumcircuit.QuantumCircuit
        Quantum circuit instance
    where_barriers: int
        Apply a barrier every where_barriers layers.

    Returns
    -------
    layered_circ: strawberryfields.program.Program or qiskit.circuit.quantumcircuit.QuantumCircuit
        Ordered circuit program with barriers
    """

    if isinstance(circ, QuantumCircuit):
        layered_circ = _to_layered_circ_qk(circ, where_barriers)
    elif SF_IS_IMPORTED and isinstance(circ, Program):
        layered_circ = _to_layered_circ_sf(circ, where_barriers)
    else:
        raise TypeError(
            "Only qiskit quantum circuit of strawberry fields programs\
             are implemented. Your circuit is of type:"
            + str(type(circ))
        )

    return layered_circ



# This code is part of qmatchatea.
#
# This code is licensed under the Apache License, Version 2.0. You may
# obtain a copy of this license in the LICENSE.txt file in the root directory
# of this source tree or at http://www.apache.org/licenses/LICENSE-2.0.
#
# Any modifications or derivative works of this code must retain this
# copyright notice, and modified files need to carry a notice indicating
# that they have been altered from the originals.

import setuptools
import os
import os.path
import importlib.util

# Parse the version file
spec = importlib.util.spec_from_file_location("qmatchatea", "./qmatchatea/version.py")
version_module = importlib.util.module_from_spec(spec)
spec.loader.exec_module(version_module)

# Parse requirements
requirement_path = "requirements.txt"
if os.path.isfile(requirement_path):
    with open(requirement_path) as fh:
        install_requires = list(fh.read().splitlines())
    install_requires.remove("mpi4py")
else:
    install_requires = []

install_requires = [
    "numpy>=1.18.1",
    "scipy>=1.4.1",
    "matplotlib>=3.1.3",
    "qtealeaves>=0.4.54,<0.5.0",
    "qiskit==0.38.0",
    "mpi4py",
    "joblib",
]

# Get the readme file
if os.path.isfile("README.md"):
    with open("README.md", "r") as fh:
        long_description = fh.read()
else:
    long_description = ""

setuptools.setup(
    name="qmatchatea",
    version=version_module.__version__,
    author="Marco Ballarin",
    author_email="marco97.ballarin@gmail.com",
    description="Quantum matcha TEA python library for tensor network emulation of quantum circuits.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://baltig.infn.it/quantum_matcha_tea/py_api_quantum_matcha_tea",
    project_urls={},
    classifiers=[
        "Programming Language :: Python :: 3",
        "Operating System :: OS Independent",
    ],
    package_dir={
        "qmatchatea": "qmatchatea",
        "qmatchatea.circuit": "qmatchatea/circuit",
    },
    packages=["qmatchatea", "qmatchatea.circuit"],
    python_requires=">=3.8",
    install_requires=install_requires,
    # entry_points = { 'console_scripts': ['build_exec = qmatchatea.bin.compiler:main', ], },
    # These packages are not mandatory for a pip installation. If they are not there
    # they will simply be ignored.
    package_data={"qmatchatea": ["bin/qmatchatea.exe", "bin/par_qmatchatea.exe"]},
)
